import os
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
from matplotlib.widgets import Button
from matplotlib.ticker import (MaxNLocator, AutoMinorLocator)
import gc
import calcdiagram_functions as calc
import logconfig as log
import data_processing as dp
import auxiliary as aux
myLogger = log.antSysDiagLogger.logger

# Basic directory path
basicDir = aux.basicDir

# Output directory path. Plotted graph images are saved here.
outputDir = os.path.join(basicDir,'output')

# legend name
legFigTitle = 'legend.png'

# Graph meta data
class GraphData:
    def __init__(self):
        self.graphTitle        = ''
        self.txAnt             = 0
        self.rxCh              = 0
        self.sensorId          = ''
        self.diagramNr         = 0
        self.figureTitle       = ''
        self.figureTitle3graph = ''
        self.antennaType       = ''
        self.nofRxCh           = 0
        self.seqNum            = 0
        self.dateAndTime       = ''
        self.frequency         = 0

        self.phyDoppGenDist          = 0
        self.doppGenHeightPtuZero    = 0
        self.rfBoardMidHeightPtuZero = 0
        self.sensorUpsideDown        = 0
        self.antennaDiagramType      = ''
        self.oppositeToPanAngle      = 0


# Save the plotted graph figure with a given title.
def saveFig(fig, figureTitle, dpi):
    figTitle = figureTitle+'.png'
    fig.set_figheight(9)
    fig.set_figwidth(16)
    figurePath = os.path.join(outputDir, figTitle)
    try:
        fig.savefig(figurePath, dpi=dpi)
    except Exception as e:
        myLogger.error(e)
    # # fig.savefig('mat.png',bbox_inches='tight')
    plt.close(fig)
    gc.collect()

# Save the plotted graph figure with a given title.
def saveFigRx(fig, figureTitle, dpi, graph, isTrgLayout):
    figTitle = figureTitle+'.png'
    fig.set_figheight(9)
    fig.set_figwidth(16)
    figurePath = os.path.join(outputDir, figTitle)
    try:
        fig.savefig(figurePath, dpi=dpi)
    except Exception as e:
        myLogger.error(e)

    if 'iff' in graph and isTrgLayout == 1:
        figTitle = figureTitle + '_3graph.png'
        fig.set_figheight(9)
        fig.set_figwidth(9)

    figurePath = os.path.join(outputDir, figTitle)
    try:
        fig.savefig(figurePath, dpi=dpi)
        myLogger.info('Image saved: '+ figurePath)
    except Exception as e:
        myLogger.error(e)    
    # fig.savefig('mat.png',bbox_inches='tight')
    plt.close(fig)
    gc.collect()


# def generateIdealCurvesData(parseDataList, desiredDiagList, dateTimeSensorList, allOnOneSlideGraphList, plotSetup):
#     angleDataList =getAngleData(parseDataList, desiredDiagList, plotSetup)
# #     # # Adding the ideal phase difference line to the phase difference graphs
# #     # if plotSetup['idealPhaseDiff'] == 1 and allOnOneSlideGraphList[graphIdx].endswith('Diff') and 'avgDiffPhase' not in allOnOneSlideGraphList[graphIdx] and 'diffPhase' not in allOnOneSlideGraphList[graphIdx]:
# #     #
# #     angleTemp = 0
# #
# #     angleTempPdIdxDict = {
# #         'angletemp': 0,
# #         'pdIdx': 0
# #     }
# #
#     angleList = []
# #     ptuList=[]
# #     angleTempList = []
#     cond = 0
# #
#     for pdIdx in range(len(parseDataList)):
#
#         for diagIdx in range(len(desiredDiagList)):
# #
# #             # for dtsIdx in range(len(dateTimeSensorList)):
# #             #
# #             #
#             if desiredDiagList[diagIdx] == parseDataList[pdIdx].fileTitle.diagram :
# #                     # dateTimeSensorList[dtsIdx]['sensor'] == parseDataList[pdIdx].fileTitle.sensor and \
# #                     # dateTimeSensorList[dtsIdx]['dateAndTime'] == parseDataList[pdIdx].fileTitle.dateAndTime:
# #
#                 # Get the ptu angle for the specific PTU2 file
#                 angle = getAngle(parseDataList, pdIdx)
#
#                 # Get the ptu angle for the specific PTU2 file
#                 tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)
#
#                 if plotSetup['targetAngle'] == 1:
#                     angle = tarAngle
#
#
#                 if cond == 0:
#                     # angleTempList.append(angle)
#                     angleTempPdIdxDict = dict()
#                     angleTempPdIdxDict.update({'angletemp': angle, 'pdIdx': pdIdx})
#
#                     angleList.append(angle)
#                     # ptuList.append(pdIdx)
#
#                     cond = 1
#                 else:
#                     for aTmpIdx in range(len(angleList)):
#                         if np.array_equal(angleList[aTmpIdx], angle) == False:
#                             # angleTempList.append(angle)
#
#                             # angleTemp = angle.copy()
#
#                             angleTempPdIdxDict = dict()
#                             angleTempPdIdxDict.update({'angletemp': angle, 'pdIdx': pdIdx})
#
#                             #angleTempPdIdxDict['angletemp'] = angleTemp
#                             # angleTempPdIdxDict['pdIdx'] = pdIdx
#
#                             angleList.append(angle)
#                             # ptuList.append(pdIdx)
#
#     return angleList

def getAngleData(parseDataList, desiredDiagList, plotSetup):
    combDictList= []
    outputList  = []
    outputDist  = dict()
    for pdIdx in range(len(parseDataList)):

        for diagIdx in range(len(desiredDiagList)):

            if desiredDiagList[diagIdx] == parseDataList[pdIdx].fileTitle.diagram:

                # Get the ptu angle for the specific PTU2 file
                angle = getAngle(parseDataList, pdIdx)

                # Get the ptu angle for the specific PTU2 file
                tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)

                if plotSetup['targetAngle'] == 1:
                    angle = tarAngle

                combDict = dict()

                firstPoint = round(angle[0], 1)
                lastPoint = round(angle[1], 1)
                step = abs(firstPoint - lastPoint)

                # step=getAngleStep(parseDataList[pdIdx])
                nofAnglePoints=len(angle)

                # angle = calc.getPtuAngVals(parseDataList)
                refAnglePoint = round(angle[0], 1)

                if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound==1:
                    diagType=parseDataList[pdIdx].measSetupInfo.antennaDiagramType
                else:
                    diagType ='NoDiagType'

                combDict.update({'nofAnglePoints': nofAnglePoints, 'angleStepSize': step,'refAnglePoint':refAnglePoint, 'angle':angle,'ptuIdx':pdIdx})
                combDictList.append(combDict)

    # outputList=list(np.unique(np.array(combDictList)))

    for pdIdx in range(len(combDictList)):
        if pdIdx==0:
            # outputDict= dict()
            # outputDict.update({'angleArray':angle,'pdIdx':pdIdx})
            # outputList.append(outputDict)
            outputList.append(combDictList[pdIdx])
            # myLogger.debug('Angle setup comb. added: '+ 'AntTyp-'+str(combDictList[pdIdx]['antType']) + '_DiagType='+combDictList[pdIdx]['diagType']+ '_NofAngPts=' + str(combDictList[pdIdx]['nofAnglePoints']) + '_AngStepSize=' + \
            #               str(combDictList[pdIdx]['angleStepSize']) + '_RefAngPt=' + str(combDictList[pdIdx]['refAnglePoint']))
        isDifferentList = []
        for outputIdx in range(len(outputList)):
            isDifferent = 0
            if outputList[outputIdx]['nofAnglePoints']!=combDictList[pdIdx]['nofAnglePoints'] or outputList[outputIdx]['angleStepSize']!=combDictList[pdIdx]['angleStepSize']\
                    or outputList[outputIdx]['refAnglePoint']!=combDictList[pdIdx]['refAnglePoint']:
               isDifferent = 1
            isDifferentList.append(isDifferent)

        found=1
        for difIdx in range(len(isDifferentList)):
            if isDifferentList[difIdx] == 0:
                found=0
                break

        if found ==1:
            outputList.append(combDictList[pdIdx])
            # myLogger.debug('Angle setup comb. added: '+ 'AntTyp-'+str(combDictList[pdIdx]['antType']) + '_DiagType='+combDictList[pdIdx]['diagType']+  '_NofAngPts=' + str(combDictList[pdIdx]['nofAnglePoints']) + '_AngStepSize=' + str(combDictList[pdIdx]['angleStepSize']) + '_RefAngPt=' + str(combDictList[pdIdx]['refAnglePoint'])+'.')
                # outputList.append(combDictList[pdIdx])

    return outputList

# Get angle step from PTU2 parsed data, for angle setup configuration.
def getAngleStep(parseData):
    # if isTargetAngle ==1:
    #     angle=calc.getTarAngVals(parseData,upsideDown)
    # else:
    #     angle=calc.getPtuAngVals(parseData)
    angle = calc.getPtuAngVals(parseData)
    firstPoint= round(angle[0],1)
    lastPoint = round(angle[1],1)
    step = abs(firstPoint-lastPoint)
    return step

# Get number fo angle points from PTU2 parsed data, for angle setup configuration.
def getNofAnglePoints(parseData):
    # if isTargetAngle == 1:
    #     length=len(calc.getTarAngVals(parseData, upsideDown))
    # else:
    #     length=len(calc.getPtuAngVals(parseData))
    length = len(calc.getPtuAngVals(parseData))
    return length

# Get the reference angle point(the first point from the angle array) from PTU2 parsed data, for angle setup configuration.
def getRefAnglePoint(parseData):
    # if isTargetAngle ==1:
    #     angle=calc.getTarAngVals(parseData,upsideDown)
    # else:
    #     angle=calc.getPtuAngVals(parseData)
    angle = calc.getPtuAngVals(parseData)
    refAnglePoint = round(angle[0],1)
    return refAnglePoint


# Get the list of all configred graphs to be drawn.
def getGraphList(graphConfig):
    graphList = []
    for graph, set in graphConfig.items():
        if set == 1:
            graphList.append(graph)
    return graphList

# Mapping string for the graph titles
def graphStringMapping(graphTitle):
    graphString     =''
    if graphTitle   == 'level':
        graphString ='Level [dB]'
    elif graphTitle == 'avgLevel':
        graphString ='Avg. Level [dB]'
    elif graphTitle == 'normLevel':
        graphString = 'Norm. Level [dB]'
    elif graphTitle == 'avgNormLevel':
        graphString = 'Avg. Norm. Level [dB]'
    elif graphTitle == 'phaseDiff':
        graphString = 'Phase Diff. [°]'
    elif graphTitle == 'avgPhaseDiff':
        graphString = 'Avg. Phase Diff. [°]'
    elif graphTitle == 'diffPhaseDiff':
        graphString = 'Diff. of Phase Diff. [°]'
    elif graphTitle == 'avgDiffPhaseDiff':
        graphString = 'Avg. Diff. of Phase Diff. [°]'
    elif graphTitle == 'corrPhaseDiff':
        graphString = 'Corr. Phase Diff. [°]'
    return graphString

# def rxChStringMapping(graphTitle, rxCh):
#     rxChString=''
#     # Mapping string for the phase difference channels
#     # phDiffRxComb = ['RxAnt1-RxAnt0', 'RxAnt2-RxAnt1', 'RxAnt3-RxAnt2']
#     if 'haseDiff'  in graphTitle:
#         lower=str(rxCh)
#         upper=str(rxCh+1)
#         rxChString='RxAnt'+upper+'-'+'RxAnt'+lower
#     else:
#         rxChString='RxAnt'+ str(rxCh)
#     return rxChString

# Get PTU angle values
def getAngle(parseDataList, x):
    angle = calc.getPtuAngVals(parseDataList[x])
    return angle

# Get target angle values
def getTargetAngle( plotSetup, parseDataList, x):
    upsideDown = getUpsideDown(plotSetup)
    tarAngle = calc.getTarAngVals(parseDataList[x], upsideDown)
    return tarAngle

# Get the info for the upside down set up sensor, from the configuration.
def getUpsideDown( plotSetup):
    if plotSetup['upsideDown'] == 0:
        upsideDown = False
    else:
        upsideDown = True
    return upsideDown



# Get the data for the actual graph for specific graph, chirp, rx channel, and PTU2 file
def getActualGraphData(graph, rxChannel, chirp, parseDataList, parseDataIdx):
    actualDiagramData = 0
    if graph == 'level':
        actualDiagramData = calc.getAntLevel(rxChannel, chirp, parseDataList[parseDataIdx])

    elif graph == 'avgLevel':  # Average power level of chirps [dB].
        actualDiagramData = calc.getAvgAntLevel(rxChannel, parseDataList[parseDataIdx])

    elif graph == 'normLevel':  # Normalised power level [dB].
        actualDiagramData = calc.getNormAntLevel(rxChannel, chirp, parseDataList[parseDataIdx])

    elif graph == 'avgNormLevel':  # Average of normalised power levels of all chirps [dB].
        actualDiagramData = calc.getAvgNormAntLevel(rxChannel, parseDataList[parseDataIdx])

    # Phase difference diagrams
    elif graph == 'phaseDiff':  # Phase difference [°].
        actualDiagramData = calc.getPhaseDiff(rxChannel, chirp, parseDataList[parseDataIdx])

    elif graph == 'avgPhaseDiff':  # Average of the phase differences of all chirps[°].
        actualDiagramData = calc.getAvgPhaseDiff(rxChannel, parseDataList[parseDataIdx])

    elif graph == 'idealPhaseDiff':  # Ideal phase difference [°].
        actualDiagramData = calc.getIdealPhaseDiff(rxChannel, chirp, parseDataList[parseDataIdx])

    elif graph == 'diffPhaseDiff':  # Difference between one phase difference and the ideal one [°].
        actualDiagramData = calc.getPhaseDiffDiff(rxChannel, chirp, parseDataList[parseDataIdx])

    elif graph == 'avgDiffPhaseDiff':  # Average difference of between the phase differences and the ideal one of all chirps[°].
        actualDiagramData = calc.getAvgPhaseDiffDiff(rxChannel, parseDataList[parseDataIdx])

    elif graph == 'corrPhaseDiff':
        actualDiagramData = calc.getPhaseDiffOffCorr(rxChannel, chirp, parseDataList[parseDataIdx])
    return actualDiagramData

# Set plot collor for each sensor
def setColor(combIdx):
    colorList = ['b', 'r', 'g', 'm', 'k', 'c', 'orange', 'midnightblue',
                  'maroon', 'springgreen', 'purple', 'dimgrey', 'teal',
                  'yellow', 'thistle', 'lime', 'darkgoldenrod', 'rosybrown',
                  'lightcoral', 'darkred', 'teal', 'navy', 'mediumslateblue',
                  'khaki', 'indigo', 'palevioletred', 'yellowgreen',
                  'palegreen', 'olivedrab', 'tomato']

    if combIdx >= len(colorList):
        res   = combIdx%len(colorList)
        color = colorList[res]
    else:
        color = colorList[combIdx]
    return color

#  Check if diagram is calculation of average from the basic types.
def isAvgDiagram(graph):
    isAvg = False
    if graph.startswith('avg'):
        isAvg = True
    return isAvg

def getAllOnOneSlideTitle(allOnOneSlideGraphList, diagramConfig, desiredDiagramList):
    title = ''
    for i in range(len(allOnOneSlideGraphList)):
        if i==0:
            title = graphStringMapping(allOnOneSlideGraphList[0]['graphType'])
        else:
            title = title+' - '+graphStringMapping(allOnOneSlideGraphList[i]['graphType'])
    if diagramConfig['selectDiagrams'] == 0:
        title = title + ' - All diagrams'

        if diagramConfig['oddOrEven'] == 1:
            title = title + ' - Odd'
        elif diagramConfig['oddOrEven'] == 2:
            title = title + ' - Even'

    elif diagramConfig['selectDiagrams'] == 1:
        title = title + ' - Diag' + str(diagramConfig['lowerNumber']) + ' - Diag' + str(
            diagramConfig['higherNumber'])
        if diagramConfig['oddOrEven'] == 1:
            title = title + ' - Odd'
        elif diagramConfig['oddOrEven'] == 2:
            title = title + ' - Even'

    else:
        title = title + ' - Diag: '
        for i in range(len(desiredDiagramList)):
            title = title + str(desiredDiagramList[i]) + ','

    return title

def getAllOnOneSlideDbLabel(allOnOneSlideGraphList):
    title = ''
    levelList = []
    for i in range(len(allOnOneSlideGraphList)):
        if 'evel' in allOnOneSlideGraphList[i]['graphType']:
            levelList.append(allOnOneSlideGraphList[i]['graphType'])
    for i in range(len(levelList)):
        if i == 0:
            title = graphStringMapping(levelList[i])
        else:
            title = title+' - '+graphStringMapping(levelList[i])
    return title

def getAllOnOneSlideDegLabel(allOnOneSlideGraphList):
    title = ''
    phDiffList = []
    for i in range(len(allOnOneSlideGraphList)):
        if 'Diff' in allOnOneSlideGraphList[i]['graphType']:
            phDiffList.append(allOnOneSlideGraphList[i]['graphType'])
    for i in range(len(phDiffList)):
        if i == 0:
            title = graphStringMapping(phDiffList[i])
        else:
            title = title+' - '+graphStringMapping(phDiffList[i])
    return title

def useSecondAxis(allOnOneSlideGraphList):
    ret = False
    phDiffFound = False
    levelFound  = False
    for graphIdx in range(len(allOnOneSlideGraphList)):
        if 'evel' in allOnOneSlideGraphList[graphIdx]['graphType']:
            levelFound = True
        elif 'Diff' in allOnOneSlideGraphList[graphIdx]['graphType']:
            phDiffFound = True
        else:
            myLogger.error('No valid graph type for all graphs at one slide configured.')
    ret=levelFound &  phDiffFound
    return ret
    # return [True for graph in allOnOneSlideGraphList if 'evel' in graph and 'Diff' in graph ]

def getPrimAxisTitle(allOnOneSlideGraphList):
    ret = ''
    for graphIdx in range(len(allOnOneSlideGraphList)):
        if 'evel' in allOnOneSlideGraphList[graphIdx]['graphType']:
            ret = getAllOnOneSlideDbLabel(allOnOneSlideGraphList)
        elif 'Diff' in allOnOneSlideGraphList[graphIdx]['graphType']:
            ret = getAllOnOneSlideDegLabel(allOnOneSlideGraphList)
        else :
            myLogger.error('No valid graph type for all graphs at one slide configured.')
    return ret


def plotAll(plotSetup, chirpConfig, allOnOneSlideGraphList, desiredDiagList, dateTimeSensorList, parseDataList, diagramConfig, pickradius):
    # Create subplots and plot figure
    fig, ax1 = plt.subplots()
    ax1.grid(True)

    ## Horizontal axis setup
    # Invert axis for showing angle from right to left
    plt.gca().invert_xaxis()

    # Set the x-axis label
    ax1.set_xlabel('PTU Angle [°]', fontsize=20)

    if plotSetup['targetAngle'] == 1:
        # angle = tarAngle
        ax1.set_xlabel('Target Angle [°]', fontsize=20)


    # Set the x-axis scale size
    if plotSetup['xScaleSizeAuto'] == 1:
        # Streching the x axis data from end to end
        ax1.autoscale(enable=True, axis='x', tight=True)
    else:
        if dp.noDiagTypeDataFound(parseDataList)==True:
            # Streching the x axis data from end to end
            ax1.autoscale(enable=True, axis='x', tight=True)
        else:
            if dp.azDataFound(parseDataList) == True:
                ax1.set_xlim(plotSetup['xScaleSizeAz'], (-1) * plotSetup['xScaleSizeAz'])
            else:
                ax1.set_xlim(plotSetup['xScaleSizeEl'], (-1) * plotSetup['xScaleSizeEl'])


    ## Vertical axis setup
    # Check if there are both level [dB] and phase diff [°] like graphs
    if useSecondAxis(allOnOneSlideGraphList):
        # Create second axis
        ax2 = ax1.twinx()
        ax2.grid(True)

        # ax2.set_zorder(0.1)

        # Set the x-axis scale size
        if plotSetup['xScaleSizeAuto'] == 1:
            # Streching the x axis data from end to end
            ax2.autoscale(enable=True, axis='x', tight=True)
        else:
            if dp.noDiagTypeDataFound(parseDataList) == True:
                # Streching the x axis data from end to end
                ax2.autoscale(enable=True, axis='x', tight=True)
            else:
                if dp.azDataFound(parseDataList) == True:
                    ax2.set_xlim(plotSetup['xScaleSizeAz'], (-1) * plotSetup['xScaleSizeAz'])
                else:
                    ax2.set_xlim(plotSetup['xScaleSizeEl'], (-1) * plotSetup['xScaleSizeEl'])

        ax1.set_ylabel(getAllOnOneSlideDbLabel(allOnOneSlideGraphList), fontsize=20)
        ax2.set_ylabel(getAllOnOneSlideDegLabel(allOnOneSlideGraphList), fontsize=20)


    else:
        # Set primary axis label
        ax1.set_ylabel(getPrimAxisTitle(allOnOneSlideGraphList), fontsize=20)

    # Set title to the graph figure
    ax1.set_title(getAllOnOneSlideTitle(allOnOneSlideGraphList, diagramConfig, desiredDiagList), fontsize=20)

    angBiggerThan45=0


    txt = None
    graphName = ''

    def onclick(event):
        if isinstance(event.artist, Line2D):

            thisline = event.artist
            # order=thisline.get_zorder()
            # print('order: ' +str(order))
            lineLabel = thisline.get_label()
            LineLabelParts = lineLabel.split('_',2)
            ax = LineLabelParts[0]
            graph = LineLabelParts[1]
            lineLabel = LineLabelParts[2]

            xdata = thisline.get_xdata()
            ydata = thisline.get_ydata()
            ind = event.ind
            if ind.size > 1:
                ind = ind[0]

            horAlign = 'left'
            if xdata[ind] < 0:
                horAlign = 'right'

            if ('ideal' in graph):
                lineLabel = graph + ': ' + str(ydata[ind]) + ', angle[°]: ' + str(xdata[ind]) + '. '
            else:
                lineLabel = graphStringMapping(graph) + ': '  + str(ydata[ind]) + ', angle[°]: ' + str(xdata[ind]) + ', ' + lineLabel
            # lineLabel = graphStringMapping(graphName) + ': ' + str(ydata[ind]) +', angle[°]: '+str(xdata[ind])+', ' + lineLabel
            if(ax=='ax1'):
                txt = ax1.text(xdata[ind], ydata[ind], lineLabel, ha=horAlign, style='italic', bbox={'facecolor': 'white', 'alpha': 0.7, 'pad': 5})
                ax1.figure.canvas.draw()
                txt.remove()
            else:
                txt = ax2.text(xdata[ind], ydata[ind], lineLabel, ha=horAlign, style='italic',bbox={'facecolor': 'white', 'alpha': 0.7, 'pad': 5})
                ax2.figure.canvas.draw()
                txt.remove()


    def toggleAxis(event):
        # print('before')
        # a1 = ax1.get_zorder()
        # print(str(a1))
        # a2 = ax2.get_zorder()
        # print(str(a2))
        # ba = btnAxis.get_zorder()
        # print(str(ba))

        if ax1.get_zorder() ==0:

            ax1.set_zorder(0.1)
            ax1.patch.set_alpha(0.5)
            ax2.set_zorder(0)
        else:
            ax2.set_zorder(0.1)
            # ax2.patch.set_alpha(0.5)
            ax1.set_zorder(0)
        # print('after')
        # a1 = ax1.get_zorder()
        # print(str(a1))
        # a2 = ax2.get_zorder()
        # print(str(a2))
        # ba = btnAxis.get_zorder()
        # print(str(ba))

    # xlimAx1=ax1.get_xlim()
    # xlimMin=xlimAx1[0]
    # xlimMax=xlimAx1[1]
    # ylimAx1=ax1.get_ylim()
    # ylimAx2=ax2.get_ylim()
    # ylimMin = min(ylimAx1[0],ylimAx2[0])
    # ylimMax = min(ylimAx1[1], ylimAx2[1])

    if pickradius == 1:
        fig.canvas.mpl_connect('pick_event', onclick)


    # Connect pickradius event to callback
    if pickradius == 1:
        # ip = InsetPosition(ax1, [0.5, 0.5 , 0.20, 0.10])  # posx, posy, width, height
        # ax1.set_axes_locator(ip)
        btnAxis = plt.axes([0.7, 0.00, 0.07, 0.050])
        btnAxis.grid(False)
        bToggleAxis = Button(btnAxis, 'ToggleAxis')
        bToggleAxis.on_clicked(toggleAxis)

    angleTemp=0

    angleTempPdIdxDict = {
        'angletemp': 0,
        'pdIdx': 0
    }

    angleList     = []
    angleTempList = []
    cond = 0

    for pdIdx in range(len(parseDataList)):

        for diagIdx in range(len(desiredDiagList)):

            for dtsIdx in range(len(dateTimeSensorList)):

                for graphIdx in range(len(allOnOneSlideGraphList)):
                    if desiredDiagList[diagIdx]==parseDataList[pdIdx].fileTitle.diagram and dateTimeSensorList[dtsIdx]['sensor'] == parseDataList[pdIdx].fileTitle.sensor and \
                            dateTimeSensorList[dtsIdx]['dateAndTime'] == parseDataList[pdIdx].fileTitle.dateAndTime:

                        # Get the ptu angle for the specific PTU2 file
                        angle = getAngle(parseDataList, pdIdx)

                        # Get the ptu angle for the specific PTU2 file
                        tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)

                        if plotSetup['targetAngle'] == 1:
                            angle = tarAngle

                        # Set xticks according to the angle values from files a plot contains
                        if abs(angle[0]) > 45:
                            angBiggerThan45=1

                        if angBiggerThan45 !=1:
                            ax1.set_xticks(np.arange(90, -91, -2.5), minor = True)
                            plt.xticks(np.arange(90, -91, -5))
                            
                        else:
                            ax1.set_xticks(np.arange(90, -91, -5), minor = True)
                            plt.xticks(np.arange(90, -91, -10))
                            

                        # Get the valid chirp list
                        validChirpList = dp.getValidChirp(chirpConfig, parseDataList, pdIdx)

                        # Get the num. of actual Rx channels for a given graph and a given PTU2 file
                        nofAcctualRxCh = dp.getNofActualRxChannels(allOnOneSlideGraphList[graphIdx]['graphType'], parseDataList, pdIdx)

                        # Set the color for the sensor specific plot lines
                        color = setColor(dtsIdx)

                        # Traverse through all rxChannels
                        for rxCh in range(nofAcctualRxCh):
                            graphName = allOnOneSlideGraphList[graphIdx]
                            plotLabel =allOnOneSlideGraphList[graphIdx]['graphType'] + '_' +dateTimeSensorList[dtsIdx]['dateAndTime'] + '_' + dateTimeSensorList[dtsIdx]['sensor']+ ', Diag' +str(desiredDiagList[diagIdx])+ ', ' + dp.rxChStringMapping(allOnOneSlideGraphList[graphIdx]['graphType'], rxCh)
                            # plotLabel = dp.rxChStringMapping(allOnOneSlideGraphList[graphIdx], rxCh)


                            # Processing for average function graphs
                            if isAvgDiagram(allOnOneSlideGraphList[graphIdx]['graphType']) == True:

                                # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                                actualGraphData = getActualGraphData(allOnOneSlideGraphList[graphIdx]['graphType'], rxCh, 0, parseDataList, pdIdx)
                                # Plot data
                                if useSecondAxis(allOnOneSlideGraphList):
                                    if 'evel' in allOnOneSlideGraphList[graphIdx]['graphType']:
                                        plotLabel = 'ax1_' + plotLabel
                                        ax1.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label = plotLabel)
                                    else:
                                        plotLabel='ax2_'+plotLabel
                                        ax2.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label=plotLabel)
                                else:
                                    plotLabel = 'ax1_' + plotLabel
                                    ax1.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label = plotLabel)
                            else:
                                # Traverse through the valid chirps
                                for chirp in range(len(validChirpList)):

                                    # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                                    actualGraphData = getActualGraphData(allOnOneSlideGraphList[graphIdx]['graphType'], rxCh, chirp, parseDataList, pdIdx)

                                    # Plot data
                                    if useSecondAxis(allOnOneSlideGraphList):
                                        if 'evel' in allOnOneSlideGraphList[graphIdx]['graphType']:
                                            plotLabel = 'ax1_' + plotLabel
                                            ax1.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label = plotLabel)
                                        else:
                                            plotLabel = 'ax2_' + plotLabel
                                            ax2.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label = plotLabel)
                                    else:
                                        plotLabel = 'ax1_' + plotLabel
                                        ax1.plot(angle, actualGraphData, '.-', pickradius=2, color=color, label = plotLabel)

                        # # Adding the ideal phase difference line to the phase difference graphs
                        # if plotSetup['idealPhaseDiff'] == 1 and allOnOneSlideGraphList[graphIdx].endswith('Diff') and 'avgDiffPhase' not in allOnOneSlideGraphList[graphIdx] and 'diffPhase' not in allOnOneSlideGraphList[graphIdx]:
                        #
                        #
                        #     if cond ==0:
                        #         # angleTempList.append(angle)
                        #         angleTempPdIdxDict=dict()
                        #         angleTempPdIdxDict.update({'angletemp': angle, 'pdIdx': pdIdx})
                        #
                        #         angleList.append(angleTempPdIdxDict)
                        #         cond = 1
                        #     else:
                        #         for aTmpIdx in range(len(angleList)):
                        #             if np.array_equal(angleList[aTmpIdx]['angletemp'], angle) == False:
                        #                 # angleTempList.append(angle)
                        #
                        #                 # angleTemp = angle.copy()
                        #                 angleTempPdIdxDict = dict()
                        #                 angleTempPdIdxDict.update({'angletemp': angle, 'pdIdx': pdIdx})
                        #                 # angleTempPdIdxDict['angletemp'] = angleTemp
                        #                 # angleTempPdIdxDict['pdIdx'] = pdIdx
                        #
                        #                 angleList.append(angleTempPdIdxDict)

    if plotSetup['idealPhaseDiff'] == 1:

        angleList = getAngleData(parseDataList, desiredDiagList, plotSetup)
        # angleList=idealCurvesDataList['angleList']
        # ptuList = idealCurvesDataList['ptuList']

        for angIdx in range(len(angleList)):
            # for ptuIdx in range(len(ptuList)):
            #     if angIdx == ptuIdx:

            nofAcctualRxCh= parseDataList[angleList[angIdx]['ptuIdx']].antennaInfo.nofAntennaDiffValues
            for rxCh in range(nofAcctualRxCh):
                actualGraphData = getActualGraphData('idealPhaseDiff', rxCh, 0, parseDataList,angleList[angIdx]['ptuIdx'])

                label = 'Phase Diff. [°] - ideal' + '_'
                if useSecondAxis(allOnOneSlideGraphList):
                    label = 'ax2_' + label
                    ax2.plot(angleList[angIdx]['angle'], actualGraphData, '-', color='olive',linewidth=3.5, pickradius=2, label=label)
                else:
                    label = 'ax1_' + label
                    ax1.plot(angleList[angIdx]['angle'], actualGraphData, '-', color='olive',linewidth=3.5, pickradius=2, label=label)

                        # # # Insert the ideal phase difference lines
                        # if plotSetup['idealPhaseDiff'] == 1 and allOnOneSlideGraphList[graphIdx].endswith('Diff') and 'avgDiffPhase' not in allOnOneSlideGraphList[graphIdx] and 'diffPhase' not in allOnOneSlideGraphList[graphIdx]:
                        #     actualGraphData = getActualGraphData('idealPhaseDiff', rxCh, 0, parseDataList, pdIdx)
                        #     label= graphStringMapping(allOnOneSlideGraphList[graphIdx]) + ' - ideal'+'_'
                        #     if useSecondAxis(allOnOneSlideGraphList):
                        #         label = 'ax2_' + label
                        #         ax2.plot(angle, actualGraphData, '-', color='olive', pickradius=2, linewidth=3.5, label=label)
                        #     else:
                        #         label = 'ax1_' + label
                        #         ax1.plot(angle, actualGraphData, '-', color='olive', pickradius=2, linewidth=3.5, label = label)

    # Assign figure title
    figureTitle = 'all_on_one_slide'

    if pickradius == 0:
        xlimits = plt.xlim()
        ylimits = plt.ylim()
        ylimtick = (ylimits[0] + ylimits[1])*0.5 + np.array((-0.5, 0.5)) * (ylimits[1] - ylimits[0]) * (1 + 0.2)
        plt.ylim(min(ylimtick), max(ylimtick))
        ax1.yaxis.set_major_locator(MaxNLocator(nbins=10))
        ax1.yaxis.set_minor_locator(AutoMinorLocator())
        ax1.grid(visible=True, which = 'minor', color = '#999999', linestyle = 'dashed', alpha = 0.2)
        # Save figure
        saveFig(fig, figureTitle, plotSetup['imageDpi'])
        # Close all plots
        # plt.close('all')
    else:

        plt.show()



# Generate graph for specific graph type, specific diagram num., all Rx channels together, and all sensors.
def plotDiagram(plotSetup, chirpConfig, graph, diagramNum, dateAndTimeSensorList, antType, parseDataList, dictScalingDict, lowerMarginDict, upperMarginDict, pickradius):
    if 'iff' in graph and plotSetup['threeGraphLayout']==1:
        font = 18
    else:
        font = 20
    # Instantiation of meta data object
    graphData = GraphData()

    # Assigning graph title to the meta data
    graphData.graphTitle=graph

    # Create subplots and plot figure
    fig, ax1 = plt.subplots()

    txt=None

    def onclick(event):
        if isinstance(event.artist, Line2D):

            thisline  = event.artist
            lineLabel = thisline.get_label()
            xdata = thisline.get_xdata()
            ydata = thisline.get_ydata()
            ind = event.ind
            if ind.size > 1 :
                ind =ind[0]
            # sign=''
            # if('evel' in graph):
            #     sign='dB'
            # else:
            #     sign ='°'

            horAlign='left'
            if xdata[ind] < 0:
                horAlign='right'

            lineLabel = graphStringMapping(graph) + ': ' + str(ydata[ind]) +', angle[°]: '+str(xdata[ind])+', ' + lineLabel

            txt = ax1.text(xdata[ind], ydata[ind], lineLabel, ha=horAlign, style='italic',bbox={'facecolor': 'white', 'alpha': 0.7, 'pad': 5})
            ax1.figure.canvas.draw()
            txt.remove()

    # Connect pickradius event to callback
    if pickradius == 1:
        fig.canvas.mpl_connect('pick_event', onclick)


    # Invert axis for showing angle from right to left
    plt.gca().invert_xaxis()

    # Show plot grid
    plt.grid(True)

    # Set title to the graph figure
    ax1.set_title(graphStringMapping(graph), fontsize=font)

    # Set y-axis label for specific general graph types
    ax1.set_ylabel(graphStringMapping(graph), fontsize=font)

    angBiggerThan45=0

    angleTemp=0

    angleTempPdIdxDict = {
        'angletemp': 0,
        'pdIdx': 0
    }

    angleTempList = []

    # Traversing through all PTU2 files
    for pdIdx in range(len(parseDataList)):

    # Traversing through all sensors
        for dateTimeSensorIdx in range(len(dateAndTimeSensorList)):
            # Looking for PTU file for specific sensor and specific diagram number
            if (diagramNum ==parseDataList[pdIdx].fileTitle.diagram) and (dateAndTimeSensorList[dateTimeSensorIdx]['sensor']==parseDataList[pdIdx].fileTitle.sensor) \
                    and dateAndTimeSensorList[dateTimeSensorIdx]['dateAndTime']==parseDataList[pdIdx].fileTitle.dateAndTime:

                # Filling in the meta data from the PTU2 title
                if parseDataList[pdIdx].fileTitle.sensorFound == 1:
                    graphData.sensorId = parseDataList[pdIdx].fileTitle.sensor
                if parseDataList[pdIdx].fileTitle.txAntFound == 1:
                    graphData.txAnt = parseDataList[pdIdx].fileTitle.txAnt
                if parseDataList[pdIdx].fileTitle.diagramFound == 1:
                    graphData.diagramNr = parseDataList[pdIdx].fileTitle.diagram
                if parseDataList[pdIdx].fileTitle.seqNumFound == 1:
                    graphData.seqNum = parseDataList[pdIdx].fileTitle.seqNum
                if parseDataList[pdIdx].fileTitle.dateAndTimeFound == 1:
                    graphData.dateAndTime = parseDataList[pdIdx].fileTitle.dateAndTime

                # Fill in the antenna type info
                graphData.antennaType = str(antType)

                # Filling in the meta data from the measurement setup inside the PTU2 file
                if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound == 1:
                    graphData.antennaDiagramType = parseDataList[pdIdx].measSetupInfo.antennaDiagramType
                if parseDataList[pdIdx].measSetupInfo.sensorUpsideDownFound == 1:
                    graphData.sensorUpsideDown = parseDataList[pdIdx].measSetupInfo.sensorUpsideDown
                if parseDataList[pdIdx].measSetupInfo.oppositeToPanAngleFound == 1:
                    graphData.oppositeToPanAngle = parseDataList[pdIdx].measSetupInfo.oppositeToPanAngle
                if parseDataList[pdIdx].measSetupInfo.phyDoppGenDistFound == 1:
                    graphData.phyDoppGenDist = parseDataList[pdIdx].measSetupInfo.phyDoppGenDist
                if parseDataList[pdIdx].measSetupInfo.rfBoardMidHeightPtuZeroFound == 1:
                    graphData.rfBoardMidHeightPtuZero = parseDataList[pdIdx].measSetupInfo.rfBoardMidHeightPtuZeroFound
                if parseDataList[pdIdx].measSetupInfo.doppGenHeightPtuZeroFound == 1:
                    graphData.doppGenHeightPtuZero = parseDataList[pdIdx].measSetupInfo.doppGenHeightPtuZeroFound

                # Get the ptu angle for the specific PTU2 file
                angle = getAngle(parseDataList, pdIdx)

                # Get the ptu angle for the specific PTU2 file
                tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)

                # Set graph title to the figure
                ax1.set_title(graphStringMapping(graph)+ ' - ' + 'AntTyp' + str(graphData.antennaType)+ ' - ' + 'Diag' + str(graphData.diagramNr), fontsize=font)

                # Level and  phase difference specific graph notations
                ax1.set_ylabel(graphStringMapping(graph), fontsize=font)

                # Set xticks according to the angle values from files a plot contains
                if abs(angle[0]) > 45:
                    angBiggerThan45 = 1

                if angBiggerThan45 != 1:
                    ax1.set_xticks(np.arange(90, -91, -2.5), minor = True)
                    plt.xticks(np.arange(90, -91, -5))
                    
                else:
                    ax1.set_xticks(np.arange(90, -91, -5), minor = True)
                    plt.xticks(np.arange(90, -91, -10))
                    

                # Set the x-axis scale size
                if plotSetup['xScaleSizeAuto'] == 1:
                    # Streching the x axis data from end to end
                    ax1.autoscale(enable=True, axis='x', tight=True)
                else:
                    if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound == 1:
                        if graphData.antennaDiagramType =='Az':
                            ax1.set_xlim(plotSetup['xScaleSizeAz'], (-1) * plotSetup['xScaleSizeAz'])
                        else:
                            ax1.set_xlim(plotSetup['xScaleSizeEl'], (-1) * plotSetup['xScaleSizeEl'])

                    else:
                        # Streching the x axis data from end to end
                        ax1.autoscale(enable=True, axis='x', tight=True)


                # Set the x-axis label
                ax1.set_xlabel('PTU Angle [°]', fontsize=font)
                if plotSetup['targetAngle'] == 1:
                    angle = tarAngle
                    ax1.set_xlabel('Target Angle [°]', fontsize=font)

                # Get the valid shirp list
                validChirpList = dp.getValidChirp(chirpConfig, parseDataList, pdIdx)

                # Get the num. of actual Rx channels for a given graph and a given PTU2 file
                nofAcctualRxCh = dp.getNofActualRxChannels(graph, parseDataList,pdIdx)

                # Set the color for the sensor specific plot lines
                color = setColor(dateTimeSensorIdx)

                # Traverse through all rxChannels
                for rxCh in range(nofAcctualRxCh):

                    # if nofAcctualRxCh <= getMaxNofRxCh(graphConfig,graphIdx, parseDataList):

                    plotLabel = graphData.dateAndTime + '_' + graphData.sensorId +', '+ dp.rxChStringMapping(graph, rxCh)

                    # Processing for average function graphs
                    if isAvgDiagram(graph) == True:
                        # # Get the actual graph for specific Rx channel, all chirps, and a specific PTU2 file
                        # actualGraphData = getActualGraphData(graph, rxCh, 0, parseDataList, pdIdx)
                        # # Plot data
                        # ax1.plot(angle, actualGraphData,'.-',color=color)
                        # Processing for diffPhaseDiff diagrams
                        if 'DiffPhaseDiff' in graph:

                            # Zoom the y-axis
                            scaleMax = dictScalingDict[diagramNum][graph]
                            # scaleMax = scaleMax[graph]

                            yAxisRange = scaleMax *  float(plotSetup['diffPhaseDiffScalingFactor'])

                            plt.ylim(-yAxisRange, yAxisRange)

                            # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                            actualGraphData = getActualGraphData(graph, rxCh, 0, parseDataList, pdIdx)
                            # Plot data
                            ax1.plot(angle, actualGraphData,  '.-',pickradius =2, color=color,label =plotLabel)

                        else:
                            # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                            actualGraphData = getActualGraphData(graph, rxCh, 0, parseDataList, pdIdx)

                            # Scale the y-axis according to the upper and lower margins of all graphs from one graph type
                            lowerMargin = lowerMarginDict[graph]
                            upperMargin = upperMarginDict[graph]
                            plt.ylim(lowerMargin, upperMargin)

                            # Plot data
                            ax1.plot(angle, actualGraphData, '.-', pickradius=2, color=color,label =plotLabel)
                    else:
                        # Traverse through the valid chirps
                        for chirp in range(len(validChirpList)):
                            # Processing for diffPhaseDiff diagrams
                            if graph.startswith('diff'):

                                # Zoom the y-axis
                                scaleMax = dictScalingDict[diagramNum][graph]

                                yAxisRange = scaleMax * float(plotSetup['diffPhaseDiffScalingFactor'])

                                plt.ylim(-yAxisRange, yAxisRange)

                                # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                                actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)
                                # Plot data
                                ax1.plot(angle, actualGraphData, '.-',pickradius=2, color=color, label=plotLabel)

                            else:

                                # Get the actual graph data for specific Rx channel, chirp and a PTU2 file
                                actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)

                                # Scale the y-axis according to the upper and lower margins of all graphs from one graph type
                                lowerMargin = lowerMarginDict[graph]
                                upperMargin = upperMarginDict[graph]
                                plt.ylim(lowerMargin, upperMargin)

                                # Plot data
                                ax1.plot(angle, actualGraphData, '.-', pickradius = 2, color=color, label =plotLabel)

                                # myLogger.info('Line plotted.')

                # Adding the ideal phase difference line to the phase difference graphs
                if graph.endswith('Diff') and 'avgDiffPhase' not in graph and 'diffPhase' not in graph and \
                        plotSetup['idealPhaseDiff'] == 1:

                    if np.array_equal(angleTemp, angle) == False:
                        angleTemp = angle
                        angleTempPdIdxDict['angletemp'] = angle
                        angleTempPdIdxDict['pdIdx'] = pdIdx

                        angleTempList.append(angleTempPdIdxDict)

    for angTempIdx in range(len(angleTempList)):
        for rxCh in range(nofAcctualRxCh):
            actualGraphData = getActualGraphData('idealPhaseDiff', rxCh, 0, parseDataList, angleTempList[angTempIdx]['pdIdx'])
            ax1.plot(angleTempList[angTempIdx]['angletemp'], actualGraphData, '-', color='olive', linewidth=3.5, pickradius=2, label='ideal curve')


                    # # Insert the ideal phase difference lines
                    # if graph.endswith('Diff') and 'avgDiffPhase' not in graph and 'diffPhase' not in graph and plotSetup['idealPhaseDiff'] == 1:
                    #     actualGraphData = getActualGraphData('idealPhaseDiff', rxCh, 0, parseDataList, pdIdx)
                    #     ax1.plot(angle, actualGraphData, '-', color='olive', pickradius = 2, linewidth=3.5,label ='ideal curve')

    # Assign figure title
    figureTitle = graphData.graphTitle  +'_AntTyp'+ str(antType)+'_TxAnt'+str(graphData.txAnt) +'_Diag'+str(graphData.diagramNr)
    graphData.figureTitle = figureTitle + '.png'

    if pickradius == 0:
        if 'iff' in graph:
            xlimits = plt.xlim()
            ylimits = plt.ylim()
            ylimtick = (ylimits[0] + ylimits[1])*0.5 + np.array((-0.5, 0.5)) * (ylimits[1] - ylimits[0]) * (1 + 0.2)
            plt.ylim(min(ylimtick), max(ylimtick))
            ax1.yaxis.set_major_locator(MaxNLocator(nbins=10))
        ax1.yaxis.set_minor_locator(AutoMinorLocator())
        ax1.grid(visible=True, which = 'minor', color = '#999999', linestyle = 'dashed', alpha = 0.2)
        # Save figure
        saveFig(fig, figureTitle, plotSetup['imageDpi'])

        #myLogger.info('Image saved')
        # Close all plots
        # plt.close('all')
    else:
        plt.show()

    #  Return graph data pbject
    return graphData

# Generate graphs for specific graph type, specific diagram num., all Rx channels separate, and all sensors.
def plotDiagramRxCh(plotSetup, chirpConfig, graph, diagramNum, rxCh, dateTimeSensorCombList, antType, parseDataList, dictScalingDict, lowerMarginDict, upperMarginDict, pickradius):
    if 'iff' in graph and plotSetup['threeGraphLayout'] == 1:
        font = 18
    else:
        font = 20
    # Instantiation of meta data object
    graphData = GraphData()

    # Assigning graph title to the meta data object
    graphData.graphTitle = graph

    # Create subplots and plot figure
    fig, ax1 = plt.subplots()

    txt=None

    def onclick(event):
        if isinstance(event.artist, Line2D):

            thisline  = event.artist
            lineLabel = thisline.get_label()
            xdata = thisline.get_xdata()
            ydata = thisline.get_ydata()
            ind   = event.ind
            if ind.size > 1 :
                ind =ind[0]

            horAlign='left'
            if xdata[ind] < 0:
                horAlign='right'

            lineLabel = graphStringMapping(graph) + ': ' + str(ydata[ind]) +', angle[°]: '+str(xdata[ind])+', ' + lineLabel
            txt = ax1.text(xdata[ind], ydata[ind], lineLabel, ha=horAlign, style='italic', bbox={'facecolor': 'white', 'alpha': 0.7, 'pad': 5})
            ax1.figure.canvas.draw()
            txt.remove()

    # Connect pickradius event to callback
    if pickradius == 1:
        fig.canvas.mpl_connect('pick_event', onclick)



    # Invert axis for showing angle from right to left
    plt.gca().invert_xaxis()
    # Show plot grid
    plt.grid(True)

    angBiggerThan45 = 0
    # ax1.grid(which='both', axis='both', linewidth=0.75, linestyle='-', color='0.75')
    angleTemp = 0

    angleTempList = []

    angleTempPdIdxDict = {
        'angletemp':0,
        'pdIdx':0
    }
    # Traversing through all PTU2 files
    for pdIdx in range(len(parseDataList)):

        # if parseDataList.parseDataList[pdIdx].antennaInfo.antennaType ==antType:
        # Traversing through all sensors
        for dateTimeSensorIdx in range(len(dateTimeSensorCombList)):
            # Looking for PTU file for specific sensor and specific diagram number
            if diagramNum == parseDataList[pdIdx].fileTitle.diagram and \
                    dateTimeSensorCombList[dateTimeSensorIdx]['sensor'] == parseDataList[pdIdx].fileTitle.sensor and \
                    dateTimeSensorCombList[dateTimeSensorIdx]['dateAndTime']==parseDataList[pdIdx].fileTitle.dateAndTime:

                # Filling in the meta data from the PTU2 title
                if parseDataList[pdIdx].fileTitle.sensorFound == 1:
                    graphData.sensorId = parseDataList[pdIdx].fileTitle.sensor
                if parseDataList[pdIdx].fileTitle.txAntFound == 1:
                    graphData.txAnt = parseDataList[pdIdx].fileTitle.txAnt
                if parseDataList[pdIdx].fileTitle.diagramFound == 1:
                    graphData.diagramNr = parseDataList[pdIdx].fileTitle.diagram
                if parseDataList[pdIdx].fileTitle.dateAndTimeFound == 1:
                    graphData.dateAndTime = parseDataList[pdIdx].fileTitle.dateAndTime

                if parseDataList[pdIdx].fileTitle.seqNumFound == 1:
                    graphData.seqNum = parseDataList[pdIdx].fileTitle.seqNum

                # Filling in the meta data from the measurement setup inside the PTU2 file
                if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound == 1:
                    graphData.antennaDiagramType = parseDataList[pdIdx].measSetupInfo.antennaDiagramType
                if parseDataList[pdIdx].measSetupInfo.sensorUpsideDownFound == 1:
                    graphData.sensorUpsideDown = parseDataList[pdIdx].measSetupInfo.sensorUpsideDown
                if parseDataList[pdIdx].measSetupInfo.oppositeToPanAngleFound == 1:
                    graphData.oppositeToPanAngle = parseDataList[pdIdx].measSetupInfo.oppositeToPanAngle
                if parseDataList[pdIdx].measSetupInfo.phyDoppGenDistFound == 1:
                    graphData.phyDoppGenDist = parseDataList[pdIdx].measSetupInfo.phyDoppGenDist
                if parseDataList[pdIdx].measSetupInfo.rfBoardMidHeightPtuZeroFound == 1:
                    graphData.rfBoardMidHeightPtuZero = parseDataList[pdIdx].measSetupInfo.rfBoardMidHeightPtuZeroFound
                if parseDataList[pdIdx].measSetupInfo.doppGenHeightPtuZeroFound == 1:
                    graphData.doppGenHeightPtuZero = parseDataList[pdIdx].measSetupInfo.doppGenHeightPtuZeroFound

                # Fill in antenna type
                graphData.antennaType = str(antType)

                # Fill in Rx channel number
                graphData.rxCh = rxCh

                # Get the PTU angle for the specific PTU2 file
                angle = getAngle(parseDataList, pdIdx)

                # Get the target angle for the specific PTU2 file
                tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)

                # Set graph title to the figure
                # ax1.set_title(graphStringMapping(graph)+ ' - ' + 'AntTyp' + str(antType) + ' - ' + 'Diag' + str(graphData.diagramNr) + ' - ' + dp.rxChStringMapping(graph, rxCh), fontsize=font)
                ax1.set_title(graphStringMapping(graph)+ ' - ' + 'AntTyp' + str(antType) + ' - ' + 'Diag' + str(graphData.diagramNr) + ' - ' + r'$\bf{' + str(dp.rxChStringMapping(graph, rxCh)) + '}$', fontsize=font)

                # Level and  phase difference specific graph notations
                ax1.set_ylabel(graphStringMapping(graph), fontsize=font)

                # Set xticks according to the angle values from files a plot contains
                if abs(angle[0]) > 45:
                    angBiggerThan45 = 1

                if angBiggerThan45 != 1:
                    ax1.set_xticks(np.arange(90, -91, -2.5), minor = True)
                    plt.xticks(np.arange(90, -91, -5)) 
                else:
                    ax1.set_xticks(np.arange(90, -91, -5), minor = True)
                    plt.xticks(np.arange(90, -91, -10))
                    

                # Set the x-axis scale size
                if plotSetup['xScaleSizeAuto'] == 1:
                    # Streching the x axis data from end to end
                    ax1.autoscale(enable=True, axis='x', tight=True)
                else:
                    if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound == 1:
                        if graphData.antennaDiagramType == 'Az':
                            ax1.set_xlim(plotSetup['xScaleSizeAz'], (-1) * plotSetup['xScaleSizeAz'])
                        else:
                            ax1.set_xlim(plotSetup['xScaleSizeEl'], (-1) * plotSetup['xScaleSizeEl'])

                    else:
                        # Streching the x axis data from end to end
                        ax1.autoscale(enable=True, axis='x', tight=True)

                # Set x-axis label
                ax1.set_xlabel('PTU Angle [°]', fontsize=font)
                if plotSetup['targetAngle'] == 1:
                    angle = tarAngle
                    ax1.set_xlabel('Target Angle [°]', fontsize=font)

                # Get the list of valid chirps
                validChirpList = dp.getValidChirp(chirpConfig, parseDataList, pdIdx)

                # Get the num. of actual Rx channels for the diagram in question and the given PTU2 file (sensor).
                nofAcctualRxCh = dp.getNofActualRxChannels(graph, parseDataList, pdIdx)

                # Set color for the sensor specific plot lines
                color = setColor(dateTimeSensorIdx)

                # Set label for each plot line
                lineLabel = graphData.dateAndTime + '_' + graphData.sensorId

                # Processing for graphs that are NOT average function.
                if isAvgDiagram(graph) == True:
                    # Check if it is a diffPhaseDiff graph
                    if 'DiffPhaseDiff' in graph:

                        # Zoom the y-axis
                        scaleMax = dictScalingDict[diagramNum][graph]
                        # scaleMax=scaleMax[graph]
                        yAxisRange = scaleMax * float(plotSetup['diffPhaseDiffScalingFactor'])

                        plt.ylim(-yAxisRange, yAxisRange)

                        # Get the actual graph values
                        actualGraphData = getActualGraphData(graph, rxCh, 0, parseDataList, pdIdx)


                        # Plot the graph
                        ax1.plot(angle, actualGraphData, '.-', color=color, pickradius = 1, label=lineLabel)

                    else:
                        # Get the actual graph values
                        actualGraphData = getActualGraphData(graph, rxCh, 0, parseDataList, pdIdx)

                        # Set the y-axis margins accordingly to the margins of all diagrams for the specific graph type
                        lowerMargin = lowerMarginDict[graph]
                        upperMargin = upperMarginDict[graph]
                        plt.ylim(lowerMargin, upperMargin)

                        # Plot the graph
                        ax1.plot(angle, actualGraphData, '.-', color=color, pickradius = 2, label=lineLabel)

                else:
                    # Traverse through the valid chirps
                    for chirp in range(len(validChirpList)):

                        # Check if it is a diffPhaseDiff graph
                        if graph.startswith('diff'):

                            # Zoom the y-axis
                            scaleMax = dictScalingDict[diagramNum][graph]
                            # scaleMax = scaleMax[graph]

                            yAxisRange = scaleMax* float(plotSetup['diffPhaseDiffScalingFactor'])

                            plt.ylim(-yAxisRange,yAxisRange)


                            # Get the actual graph values
                            actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)

                            # Plot the graph
                            ax1.plot(angle, actualGraphData,'.-', color=color, pickradius = 2, label=lineLabel)

                        else:
                            # Get the actual graph values
                            actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)

                            # Set the y-axis margins accordingly to the margins of all diagrams for the specific graph type
                            lowerMargin = lowerMarginDict[graph]
                            upperMargin = upperMarginDict[graph]
                            plt.ylim(lowerMargin, upperMargin)

                            # Plot the graph
                            ax1.plot(angle, actualGraphData,  '.-', color=color, pickradius = 2, label=lineLabel)

                            # myLogger.info('line plotted.')

                # Adding the ideal phase difference line to the phase difference graphs
                if graph.endswith('Diff') and 'avgDiffPhase' not in graph and 'diffPhase' not in graph and plotSetup['idealPhaseDiff'] == 1:

                    if np.array_equal(angleTemp,angle) == False:

                        angleTemp = angle
                        angleTempPdIdxDict['angletemp'] = angle
                        angleTempPdIdxDict['pdIdx'] = pdIdx

                        angleTempList.append(angleTempPdIdxDict)

    for angTempIdx in range(len(angleTempList)):

        actualGraphData = getActualGraphData('idealPhaseDiff', rxCh, 0, parseDataList, angleTempList[angTempIdx]['pdIdx'])
        ax1.plot(angleTempList[angTempIdx]['angletemp'], actualGraphData, '-', color='olive', linewidth=3.5, pickradius=2, label='ideal curve')

    # Asigning figure title and puting it in the graph meta data object
    figureTitle = graphData.graphTitle +'_AntTyp'+ str(antType) +'_TxAnt'+str(graphData.txAnt) +'_Diag'+str(graphData.diagramNr) +'_RxCh'+str(graphData.rxCh)
    graphData.figureTitle = figureTitle + '.png'
    if 'iff' in graph and plotSetup['threeGraphLayout'] == 1:
        graphData.figureTitle3graph = figureTitle+'_3graph.png'

    if pickradius == 0:
        if 'iff' in graph:
            xlimits = plt.xlim()
            ylimits = plt.ylim()
            ylimtick = (ylimits[0] + ylimits[1])*0.5 + np.array((-0.5, 0.5)) * (ylimits[1] - ylimits[0]) * (1 + 0.2)
            plt.ylim(min(ylimtick), max(ylimtick))
            ax1.yaxis.set_major_locator(MaxNLocator(nbins=10))
        ax1.yaxis.set_minor_locator(AutoMinorLocator())
        ax1.grid(visible=True, which = 'minor', color = '#999999', linestyle = 'dashed', alpha = 0.2)
        # Save figure
        saveFigRx(fig, figureTitle, plotSetup['imageDpi'], graph, plotSetup['threeGraphLayout'])

        #myLogger.info('Image saved')
        # Close all plots
        # plt.close('all')


    else:
        plt.show()

    # Return the meta data object
    return graphData

# Generate legend figure for all sensors
def generateLegend(plotSetup, dateTimeSensorCombList):

    # Create legend figure
    legFig = plt.figure(figsize=(2, 6))

    # plt.figure()

    # Num. of sensors
    nofComb = len(dateTimeSensorCombList)
    # Maximal possible number of sensors
    if plotSetup['idealPhaseDiff'] == 1:
        maxNofComb = 16

    else:
        maxNofComb = 17

    # Max num of lines
    maxNofLines = 18

    # Create pahtch list
    patchList  = []
    whiteBlank ='white'

    if nofComb>maxNofComb:
        # Insert sensor list fith specific color patch
        for combIdx in range(maxNofComb):
            color = setColor(combIdx)
            patchList.append(mpatches.Patch(color=color, label=dateTimeSensorCombList[combIdx]['dateAndTime'] + '_' +dateTimeSensorCombList[combIdx]['sensor']))
        # Add the ideal phase difference line
        if plotSetup['idealPhaseDiff'] == 1:
            patchList.append(mpatches.Patch(color='olive', label='Ideal phase-diff'))
    else:
        if plotSetup['idealPhaseDiff'] == 1:
            nofSensorLines = nofComb+1
        else:
            nofSensorLines = nofComb
        nofEmptyLines = int((maxNofLines-nofSensorLines)/2)
        for i in range(nofEmptyLines):
            # Add empty space for better formating
            patchList.append(mpatches.Patch(color=whiteBlank, label=''))
        for combIdx in range(nofComb):
            color = setColor(combIdx)
            patchList.append(mpatches.Patch(color=color, label=dateTimeSensorCombList[combIdx]['dateAndTime'] + '_' +dateTimeSensorCombList[combIdx]['sensor']))

        # Add the ideal phase difference line
        if plotSetup['idealPhaseDiff'] == 1:
            patchList.append(mpatches.Patch(color='olive', label='Ideal phase-diff'))
        for i in range(nofEmptyLines):
            # Add empty space for better formating
            patchList.append(mpatches.Patch(color=whiteBlank, label=''))

    # Attach patch list to a legend
    legFig.legend(handles=patchList, loc='center')


    # Figure title and save path
    legFigPath = os.path.join(outputDir, legFigTitle)
    # Save figure
    try:
        legFig.savefig(legFigPath, bbox_inches='tight', pad_inches=0, dpi=150)
    except Exception as e:
        myLogger.error(e)
    #
    # if plotSetup['ShowCertainPlotWindow']==1:
    #     plt.show()


# Lower margin of all drawn sensor diagrams for a specific graph
def scaleLowerMargin( graph, chirpSetup, parseDataList):
    actualGraph = 0
    lowerMarginList = []
    for pdIdx in range(len(parseDataList)):

        validChirpList = dp.getValidChirp(chirpSetup, parseDataList,pdIdx)
        nofRxCh = dp.getMaxNofRxCh(graph, parseDataList)
        for chirp in range(len(validChirpList)):
            for rxCh in range(nofRxCh):
                actualGraph = getActualGraphData(graph,rxCh,chirp,parseDataList,pdIdx)
                minValue = actualGraph.min()
                lowerMarginList.append(minValue)

    if len(lowerMarginList) !=0:
        lowerMargin = min(lowerMarginList)
        whole = int(abs(lowerMargin) / 10)
        rest = abs(lowerMargin) % 10

        if lowerMargin >= 0:
            if rest>5:
                lowerMargin = whole*10
            else:
                lowerMargin = whole*10-10
        else:
            if rest <5:
                lowerMargin = whole*(-10)-10
            else:
                lowerMargin = whole*(-10)-20
        return lowerMargin

# Upper margin of all drawn sensor diagrams for a specific graph
def scaleUpperMargin(graph, chirpSetup, parseDataList):
    actualGraph = 0
    upperMarginList = []
    for pdIdx in range(len(parseDataList)):

        validChirpList = dp.getValidChirp(chirpSetup, parseDataList,pdIdx)
        nofRxCh = dp.getMaxNofRxCh(graph, parseDataList)
        for chirp in range(len(validChirpList)):
            for rxCh in range(nofRxCh):
                actualGraph = getActualGraphData(graph,rxCh,chirp,parseDataList,pdIdx)
                maxValue = actualGraph.max()
                upperMarginList.append(maxValue)

    if len(upperMarginList) !=0:
        upperMargin = max(upperMarginList)
        whole = int(abs(upperMargin) / 10)
        rest = abs(upperMargin) % 10

        if upperMargin >= 0:
            if rest > 5:
                upperMargin = whole * 10 + 20
            else:
                upperMargin = whole * 10 + 10
        else:
            if rest < 5:
                upperMargin = whole * (-10) + 10
            else:
                upperMargin = whole *(-10)
        return upperMargin

# Scaling factor for the diffPhaseDiff and avgDifPhaseDiff diagrams
def scale(graph, chirpSetup, plotSetup, parseDataList):
    absMaxList = []
    maxValue = 0
    minValue = 0
    upperXMarginIdx = 0
    lowerXMarginIdx = 0
    # if graph=='diffPhaseDiff' or graph=='avgDiffPhaseDiff':
    for pdIdx in range(len(parseDataList)):
        # Get the PTU angle for the specific PTU2 file
        angle = getAngle(parseDataList, pdIdx)
        # Get the target angle for the specific PTU2 file
        tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)
        # Set the angle according to the configuration
        if plotSetup['targetAngle'] == 1:
            angle=tarAngle
        validChirpList = dp.getValidChirp(chirpSetup, parseDataList,pdIdx)
        nofRxCh = dp.getMaxNofRxCh(graph, parseDataList)
        for chirp in range(len(validChirpList)):
            for rxCh in range(nofRxCh):
                actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)
                arrayLen = len(angle)
                for i in range(arrayLen):
                    if angle[i] <= 10 + 0.1 and angle[i] >= 10 - 0.1:
                        if plotSetup['targetAngle'] == 1:
                            if getUpsideDown(plotSetup) == False:
                                lowerXMarginIdx = i
                            else:
                                upperXMarginIdx = i
                        else:
                            if getUpsideDown(plotSetup) == False:
                                upperXMarginIdx = i
                            else:
                                lowerXMarginIdx = i
                    if angle[i] <= -10 + 0.1 and angle[i] >= -10 - 0.1:
                        if plotSetup['targetAngle'] == 1:
                            if getUpsideDown(plotSetup) == False:
                                upperXMarginIdx = i
                            else:
                                lowerXMarginIdx = i
                        else:
                            if getUpsideDown(plotSetup) == False:
                                lowerXMarginIdx = i
                            else:
                                upperXMarginIdx = i
                maxValue = actualGraphData[lowerXMarginIdx:upperXMarginIdx].max()
                minValue = actualGraphData[lowerXMarginIdx:upperXMarginIdx].min()
                if abs(maxValue) >= abs(minValue):
                    absMax = abs(maxValue)
                else:
                    absMax = abs(minValue)
                absMaxList.append(absMax)
    return max(absMaxList)

# Scaling factor for the diffPhaseDiff and avgDifPhaseDiff diagrams
def scalePerDiagram(graph, diagramNr, chirpSetup, plotSetup, parseDataList):
    absMaxList = []
    maxValue = 0
    minValue = 0
    upperXMarginIdx = 0
    lowerXMarginIdx = 0
    # if graph=='diffPhaseDiff' or graph=='avgDiffPhaseDiff':
    for pdIdx in range(len(parseDataList)):

        # for diagramIdx in range(len(desiredDiagramList)):
        if parseDataList[pdIdx].fileTitle.diagram == diagramNr:
            # Get the PTU angle for the specific PTU2 file
            angle = getAngle(parseDataList, pdIdx)
            # Get the target angle for the specific PTU2 file
            tarAngle = getTargetAngle(plotSetup, parseDataList, pdIdx)
            # Set the angle according to the configuration
            if plotSetup['targetAngle'] == 1:
                angle = tarAngle
            validChirpList = dp.getValidChirp(chirpSetup, parseDataList, pdIdx)
            nofRxCh = dp.getMaxNofRxCh(graph, parseDataList)
            for chirp in range(len(validChirpList)):
                for rxCh in range(nofRxCh):

                    actualGraphData = getActualGraphData(graph, rxCh, chirp, parseDataList, pdIdx)
                    arrayLen = len(angle)

                    for i in range(arrayLen):
                        if plotSetup['targetAngle'] == 1:
                            if getUpsideDown(plotSetup) == False:
                                if angle[i] <= 10 + 0.1:
                                    lowerXMarginIdx=i
                                    break
                            else:
                                if angle[i] >= -10 - 0.1:
                                    lowerXMarginIdx = i
                                    break
                        else:
                            if angle[i] >= -10 - 0.1:
                                lowerXMarginIdx = i
                                break

                    for i in range(arrayLen):
                        if plotSetup['targetAngle'] == 1:
                            if getUpsideDown(plotSetup) == False:
                                if angle[i] < -10 - 0.1:
                                    upperXMarginIdx = i
                                    break
                                elif i == (len(angle)-1):
                                    upperXMarginIdx=i
                            else:
                                if angle[i] > 10 + 0.1:
                                    upperXMarginIdx =i
                                    break
                                elif i == (len(angle) - 1):
                                    upperXMarginIdx = i
                        else:
                            if angle[i] > 10 + 0.1:
                                upperXMarginIdx =i
                                break
                            elif i == (len(angle) - 1):
                                upperXMarginIdx = i

                    maxValue = actualGraphData[lowerXMarginIdx:upperXMarginIdx].max()
                    minValue = actualGraphData[lowerXMarginIdx:upperXMarginIdx].min()
                    if abs(maxValue) >= abs(minValue):
                        absMax = abs(maxValue)
                    else:
                        absMax = abs(minValue)
                    absMaxList.append(absMax)
    if len(absMaxList) !=0:
        maximum=max(absMaxList)
        return maximum



import calcdiagram_functions as calc
import logconfig as log
import numpy as np
from abc import ABC, abstractmethod
import copy

# Get logger instance
myLogger =log.antPatLogger.logger

class OutFormatTable(ABC):
    def __init__(self):
        self.antennaType=0
        self.chirp=0
        self.azElCombSetup=0
        self.dataTablesList=[]
        self.outputDict=0
        self.dataTablesDictList=[]

    # Reorder the data tables list by changing the antenna pattern index in the configured tables.
    @abstractmethod
    def reorderDataTablesList(self):
        pass

# JSON Table class. Contains parameters for the generated JSON files.
class JSON_Table(OutFormatTable):
    def __init__(self):
        super().__init__()
        self.jsonFileTitle = ''

    # Reorder the data tables list by changing the antenna pattern index in the configured tables.
    def reorderDataTablesList(self):
        myLogger.info("JSON file data tables list reordered, Antenna Type: "+str(self.antennaType) +" Diagram type and Angle setup comb.: " + self.azElCombSetup.string + ', Chirp' + str(self.chirp) + ':')

        for i in range(len(self.dataTablesList)):

            oldIdx=self.dataTablesList[i].configTable.antennaPatternIdx
            self.dataTablesList[i].configTable.antennaPatternIdx=i
            newIdx=self.dataTablesList[i].configTable.antennaPatternIdx

            myLogger.info('[ Table'+str(oldIdx)+' - "antenna_pattern_idx": '+str(oldIdx) +' ] - - - - > [ '+'Table'+str(newIdx)+' - "antenna_pattern_idx": '+str(newIdx)+' ]')

class YAML_Table(OutFormatTable):
    def __init__(self):
        super().__init__()
        self.yamlFileTitle = ''

    # Reorder the data tables list by changing the antenna pattern index in the configured tables.
    def reorderDataTablesList(self):
        myLogger.info("YAML file data tables list reordered, Antenna Type: "+str(self.antennaType) +" Diagram type and Angle setup comb.: " + self.azElCombSetup.string + ', Chirp' + str(self.chirp) + ':')

        for i in range(len(self.dataTablesList)):

            oldIdx=self.dataTablesList[i].configTable.antennaPatternIdx
            self.dataTablesList[i].configTable.antennaPatternIdx=i
            newIdx=self.dataTablesList[i].configTable.antennaPatternIdx

            myLogger.info('[ Table'+str(oldIdx)+' - "antenna_pattern_idx": '+str(oldIdx) +' ] - - - - > [ '+'Table'+str(newIdx)+' - "antenna_pattern_idx": '+str(newIdx)+' ]')


# Setup class. Contains the input parameters data for data tables.
class Setup():
    def __init__(self, setup, antPatHeader, ptu2InputConfDict, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        self.setup=0
        self.antPatHeader=0
        self.ptu2InputConf=0
        self.antennaTypesList=[]
        self.angleSetupList=[]
        self.confChirpList=[]
        self.confTablesList=[]
        self.angleSetupAzElComb=[]
        self.jsonTablesList=[]
        self.yamlTablesList=[]
        self.string=""
        self.errorStatus=0
        self.setupValues(setup, antPatHeader, ptu2InputConfDict, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData)

    # Set the list of AZ-El combinations.
    def setAngleSetupAzElComb(self,angleSetupList):
        outputList = []
        azList = []
        elList = []
        noDiagTypeList=[]
        filteredList=[]

        if len(angleSetupList)==0:
           self.errorStatus=1
           myLogger.error("Empty angle setup list!")
        else:
            for i in range(len(angleSetupList)):
                if angleSetupList[i].diagType == 'NoDiagType':
                    noDiagTypeList.append(angleSetupList[i])

            if len(noDiagTypeList)!=0:
                for antIdx in range(len(self.antennaTypesList)):
                    filteredList=getAngSetupListPerAntennaType(self.antennaTypesList[antIdx], angleSetupList)
                    if len(filteredList)!=0:
                        if len(filteredList) == 1:
                            element = AzElComb(filteredList[0], filteredList[0], 0)
                            outputList.append(element)
                        else:
                            for i in range(len(filteredList)):
                                for j in range(len(filteredList)):
                                    if filteredList[i].nofAnglePoints !=filteredList[j].nofAnglePoints or \
                                            filteredList[i].refAnglePoint != filteredList[j].refAnglePoint or \
                                            filteredList[i].angleStepSize !=filteredList[j].angleStepSize:
                                        element = AzElComb(filteredList[i], filteredList[j], 0)
                                        outputList.append(element)


            else:
                for i in range(len(angleSetupList)):
                    if angleSetupList[i].diagType=='Az':
                        azList.append(angleSetupList[i])
                for i in range(len(angleSetupList)):
                    if angleSetupList[i].diagType == 'El':
                        elList.append(angleSetupList[i])

                if len(azList)==0 and len(elList)!=0:
                    for j in range(len(elList)):
                        element = AzElComb(0, elList[j],0)
                        outputList.append(element)
                elif len(elList)==0 and len(azList)!=0:
                    for i in range(len(azList)):
                        element=AzElComb(azList[i],0,0)
                        outputList.append(element)
                elif len(elList)!=0 and len(azList)!=0:
                    for i in range(len(azList)):
                        for j in range(len(elList)):
                            element = AzElComb(azList[i], elList[j],0)
                            outputList.append(element)

        return outputList

    # Set string.
    def setStr(self,ptu2InputConf):
        self.string=ptu2InputConf.getStr()

    # Set the list of angle setups.
    def setAngleSetupList(self, ptu2InputConf, parseDataList):
        angSetupCombDictList=getAppearingAngleSetupComb(parseDataList)
        angSetupCombDictList1=[]
        withDiagTypeList=[]
        withoutDiagTypeList=[]
        angSetupCombList = []

        for i in range(len(angSetupCombDictList)):
            if angSetupCombDictList[i]["diagType"]=='Az' or angSetupCombDictList[i]["diagType"]=='El':
                withDiagTypeList.append(angSetupCombDictList[i])

        for i in range(len(angSetupCombDictList)):
            if angSetupCombDictList[i]["diagType"] == 'NoDiagType':
                withoutDiagTypeList.append(angSetupCombDictList[i])

        if ptu2InputConf.diagTypeInfo == 1:
            angSetupCombDictList1=withDiagTypeList.copy()
        else:
            angSetupCombDictList1 = withoutDiagTypeList.copy()

        if len(angSetupCombDictList1)==0:
            myLogger.error('No data for ' + ptu2InputConf.getStr() + ' found.')
            self.errorStatus=1
        else:
            for angSetupIdx in range(len(angSetupCombDictList1)):
                myLogger.info('Angle setup comb. added: '+'AntType-'+ str(+angSetupCombDictList1[angSetupIdx]["antType"])+'_DiagType='+angSetupCombDictList1[angSetupIdx]["diagType"]+ '_NofPts=' + str(angSetupCombDictList1[angSetupIdx]["nofAnglePoints"]) + '_StepSize=' + str(angSetupCombDictList1[angSetupIdx]["angleStepSize"]) + '_RefPt=' + str(angSetupCombDictList1[angSetupIdx]["refAnglePoint"])+'.')

                angSetupComb = AngleSetup(angSetupCombDictList1[angSetupIdx])
                angSetupCombList.append(angSetupComb)

        return angSetupCombList

    # Set the configured tables list.
    def setConfTablesList(self,ptu2InputConf):

        outputList = []
        for confTabIdx in range(len(ptu2InputConf.tables)):
            configTable =ConfigTable(ptu2InputConf.tables[confTabIdx])
            if configTable.errorStatus==1:
                self.errorStatus=1
            outputList.append(configTable)
        return outputList

    # Seting the class members according to the input parameters.
    def setupValues(self, setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        self.setup = setup
        self. antPatHeader=copy.deepcopy(antPatHeader)
        self.ptu2InputConf = ptu2InputConf

        if len(parseDataList) == 0:
            self.errorStatus = 1
            myLogger.error('No parsed PTU-2 files found!')

        self.antennaTypesList=getAntennaTypesList(parseDataList,self.ptu2InputConf)

        # Set error if Antenna Type list is empty.
        if len(self.antennaTypesList)==0:
            self.errorStatus=1
            myLogger.error('No antenna types found.')

        self.confChirpList = getConfChirpList(chirpConfig)

        # Set error if Chirp list is empty.
        if len(self.confChirpList)==0:
            self.errorStatus=1
            myLogger.error('Invalid chirp configuration.')

        self.angleSetupList = self.setAngleSetupList(self.ptu2InputConf, parseDataList)
        self.confTablesList = self.setConfTablesList(self.ptu2InputConf)
        self.angleSetupAzElComb=self.setAngleSetupAzElComb(self.angleSetupList)

        self.setStr(self.ptu2InputConf)

        myLogger.info("Initial number of data tables: "+str(self.antPatHeader["nofDataTables"])+'.')

        if self.ptu2InputConf.diagTypeInfo == 1:
            myLogger.info("Only PTU2 files WITH diagram type info considered.")
        else:
            myLogger.info("Only PTU2 files WITHOUT diagram type info considered.")

        if self.ptu2InputConf.upsideDownInfo == 1:
            myLogger.info("Only PTU2 files WITH upside down info considered.")
        else:
            myLogger.info("Only PTU2 files WITHOUT upside down info considered. Processing accordingly to the configured 'upsideDown': "+str(self.ptu2InputConf.upsideDown)+", value.")

        # Set error if the configured tables list surpasses the maximum data tables allowed.
        if len(self.confTablesList) > MAX_NOF_DATA_TABLES:
            self.errorStatus=1
            myLogger.error("The configured tables list surpasses the maximal number of data tables allowed: "+str(MAX_NOF_DATA_TABLES)+'.')

        # Adjust the nofDataTables field to the number of configured tables.
        elif self.antPatHeader["nofDataTables"] != len(self.confTablesList):
            self.antPatHeader["nofDataTables"]=len(self.confTablesList)
            myLogger.info("Nof. Data Tables adjusted to the number of configured tables: " + str(
                len(self.confTablesList)) + '.')

class Setup_v2(Setup):
    def __init__(self, setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        self.portData = portData
        super().__init__(setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData)


    # Seting the class members according to the input parameters.
    def setupValues(self, setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        self.setup = setup
        self.antPatHeader = copy.deepcopy(antPatHeader)
        self.portData = copy.deepcopy(portData)
        self.ptu2InputConf = ptu2InputConf

        if len(parseDataList)==0:
            self.errorStatus=1
            myLogger.error('No parsed PTU-2 files found!')

        self.antennaTypesList = getAntennaTypesList(parseDataList, self.ptu2InputConf)

        # Set error if Antenna Type list is empty.
        if len(self.antennaTypesList) == 0:
            self.errorStatus = 1
            myLogger.error('No antenna types found.')

        self.confChirpList = getConfChirpList(chirpConfig)

        # Set error if Chirp list is empty.
        if len(self.confChirpList) == 0:
            self.errorStatus = 1
            myLogger.error('Invalid chirp configuration.')

        self.angleSetupList = self.setAngleSetupList(self.ptu2InputConf, parseDataList)
        self.confTablesList = self.setConfTablesList(self.ptu2InputConf)
        self.angleSetupAzElComb = self.setAngleSetupAzElComb(self.angleSetupList)

        self.setStr(self.ptu2InputConf)

        myLogger.info("Initial number of data tables: " + str(self.portData["nofDataTables"]) + '.')

        if self.ptu2InputConf.diagTypeInfo == 1:
            myLogger.info("Only PTU2 files WITH diagram type info considered.")
        else:
            myLogger.info("Only PTU2 files WITHOUT diagram type info considered.")

        if self.ptu2InputConf.upsideDownInfo == 1:
            myLogger.info("Only PTU2 files WITH upside down info considered.")
        else:
            myLogger.info("Only PTU2 files WITHOUT upside down info considered. Processing accordingly to the configured 'upsideDown': " + str(
                    self.ptu2InputConf.upsideDown) + ", value.")

        # Set error if the configured tables list surpasses the maximum data tables allowed.
        if len(self.confTablesList) > MAX_NOF_DATA_TABLES:
            self.errorStatus = 1
            myLogger.error("The configured tables list surpasses the maximal number of data tables allowed: " + str(
                MAX_NOF_DATA_TABLES) + '.')

        # Adjust the nofDataTables field to the number of configured tables.
        elif self.portData["nofDataTables"] != len(self.confTablesList):
            self.portData["nofDataTables"] = len(self.confTablesList)
            myLogger.info("Nof. Data Tables adjusted to the number of configured tables: " + str(
                len(self.confTablesList)) + '.')

class Setup_v3(Setup_v2):
    def __init__(self, setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        super().__init__(setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData)

        # Set the list of angle setups.

    def setAngleSetupList(self, ptu2InputConf, parseDataList):
        angSetupCombDictList = getAppearingAngleSetupComb(parseDataList)
        withDiagTypeList = []
        angSetupCombList = []

        for i in range(len(angSetupCombDictList)):
            if angSetupCombDictList[i]["diagType"] == 'Az' or angSetupCombDictList[i]["diagType"] == 'El':
                withDiagTypeList.append(angSetupCombDictList[i])


        if len(withDiagTypeList) == 0:
            myLogger.error('No data for ' + ptu2InputConf.getStr() + ' found.')
            self.errorStatus = 1
        else:
            for angSetupIdx in range(len(withDiagTypeList)):
                myLogger.info('Angle setup comb. added: ' + 'AntType-' + str(
                    +withDiagTypeList[angSetupIdx]["antType"]) + '_DiagType=' + withDiagTypeList[angSetupIdx]["diagType"] + '_NofPts=' + str(withDiagTypeList[angSetupIdx]["nofAnglePoints"]) + '_StepSize=' + str(withDiagTypeList[angSetupIdx]["angleStepSize"]) + '_RefPt=' + str(withDiagTypeList[angSetupIdx]["refAnglePoint"]) + '.')

                angSetupComb = AngleSetup(withDiagTypeList[angSetupIdx])
                angSetupCombList.append(angSetupComb)

        return angSetupCombList

# Seting the class members according to the input parameters.
    def setupValues(self, setup, antPatHeader, ptu2InputConf, chirpConfig, parseDataList, MAX_NOF_DATA_TABLES,portData):
        self.setup = setup
        self.antPatHeader=copy.deepcopy(antPatHeader)
        self.portData = copy.deepcopy(portData)
        self.ptu2InputConf = ptu2InputConf

        if len(parseDataList) == 0:
            self.errorStatus = 1
            myLogger.error('No parsed PTU-2 files found!')

        self.antennaTypesList=getAntennaTypesList(parseDataList,self.ptu2InputConf)

        # Set error if Antenna Type list is empty.
        if len(self.antennaTypesList)==0:
            self.errorStatus=1
            myLogger.error('No antenna types found.')

        self.confChirpList = getConfChirpList(chirpConfig)

        # Set error if Chirp list is empty.
        if len(self.confChirpList)==0:
            self.errorStatus=1
            myLogger.error('Invalid chirp configuration.')

        self.angleSetupList = self.setAngleSetupList(self.ptu2InputConf, parseDataList)
        self.confTablesList = self.setConfTablesList(self.ptu2InputConf)
        self.angleSetupAzElComb=self.setAngleSetupAzElComb(self.angleSetupList)

        self.setStr(self.ptu2InputConf)

        myLogger.info("Initial number of data tables: "+str(self.portData["nofDataTables"])+'.')


        # Set error if the configured tables list surpasses the maximum data tables allowed.
        if len(self.confTablesList) > MAX_NOF_DATA_TABLES:
            self.errorStatus=1
            myLogger.error("The configured tables list surpasses the maximal number of data tables allowed: "+str(MAX_NOF_DATA_TABLES)+'.')

        # Adjust the nofDataTables field to the number of configured tables.
        elif self.portData["nofDataTables"] != len(self.confTablesList):
            self.portData["nofDataTables"]=len(self.confTablesList)
            myLogger.info("Nof. Data Tables adjusted to the number of configured tables: " + str(len(self.confTablesList)) + '.')



class Ptu2InputConf():
    def __init__(self,ptu2InputConf,confIdx):
        self.diagTypeInfo=0
        self.targetAngle=0
        self.upsideDown=0
        self.upsideDownInfo=0
        # self.elDiagCriteria=0
        self.confIdx=0
        self.tables=[]
        self.str=''
        self.setup(ptu2InputConf,confIdx)

    def setStr(self):
        diagTypeInfoStr=''
        targetAngleStr=''
        upsideDownStr=''
        upsideDownInfoStr=''
        if self.diagTypeInfo!=1:
            diagTypeInfoStr='NoDiagType'
        else:
            diagTypeInfoStr='DiagType'

        if self.upsideDown==1:
            upsideDownStr='_UpDown=1_'
        else:
            upsideDownStr='_UpDown=0_'

        if self.upsideDownInfo!=1:
            upsideDownInfoStr='NoUpDownInfo'
        else:
            upsideDownInfoStr='UpDownInfo'
            upsideDownStr='_'

        if self.targetAngle==0:
            targetAngleStr='PTU-Ang'
        else:
            targetAngleStr='Tar-Ang'

        self.str='Conf' +str(self.confIdx) +'_'+diagTypeInfoStr +'_'+ upsideDownInfoStr + upsideDownStr + targetAngleStr

    # Get string.
    def getStr(self):
        self.setStr()
        return self.str

    def setup(self,ptu2InputConf,confIdx):
        self.diagTypeInfo =ptu2InputConf["diagTypeInfo"]
        self.targetAngle = ptu2InputConf["targetAngle"]
        self.upsideDown = ptu2InputConf["upsideDown"]
        self.upsideDownInfo = ptu2InputConf["upsideDownInfo"]
        # self.elDiagCriteria=ptu2InputConf["elDiagCriteria"]
        self.confIdx=confIdx
        self.tables=ptu2InputConf["tables"]
        self.setStr()

class Ptu2InputConfV3():
    def __init__(self,ptu2InputConf):
        self.targetAngle=0
        self.upsideDown=0
        # self.elDiagCriteria=0
        self.tables=[]
        self.str=''
        self.setup(ptu2InputConf)

    def setStr(self):

        targetAngleStr=''
        upsideDownStr=''

        if self.upsideDown==1:
            upsideDownStr='UpDown=1'
        else:
            upsideDownStr='UpDown=0'

        if self.targetAngle==0:
            targetAngleStr='PTU-Ang'
        else:
            targetAngleStr='Tar-Ang'

        self.str='Conf' +str(self.confIdx)  +'_'+ upsideDownStr +'_'+ targetAngleStr

    # Get string.
    def getStr(self):
        self.setStr()
        return self.str

    def setup(self,ptu2InputConf,confIdx):

        self.targetAngle = ptu2InputConf["targetAngle"]
        self.upsideDown = ptu2InputConf["upsideDown"]
        self.confIdx=confIdx
        self.tables=ptu2InputConf["tables"]
        self.setStr()

# Class that contains a combination of Az and El angle setups or No diagram type info angle setup.
class AzElComb():
    def __init__(self,az,el,noDiagType):
        self.az=0
        self.el=0
        self.noDiagType=0
        self.string=''
        self.setup(az,el,noDiagType)

    # Set string.
    def setStr(self,az,el,noDiagType):
        str=''
        if noDiagType!=0:
            str=noDiagType.angleSetupStr
        else:
            if el==0:
                str=az.angleSetupStr
            elif az==0:
                str=el.angleSetupStr
            else:
                str=az.angleSetupStr +' - '+el.angleSetupStr
        return str

    # Set values according to input parameters.
    def setup(self,az,el,noDiagType):
        self.az=az
        self.el=el
        self.noDiagType=noDiagType
        self.string=self.setStr(az,el,noDiagType)

# Class containing the data for Antenna Pattern WITH diagram type info, for certain Configured Table and certain Antenna Type, Angle setup, and Chirp.
class DataTable():
    def __init__(self,angleSetupData,configTable,chirp,parseDataList,ptu2InputConf):
        self.ptu2InputConf=0
        self.tableMeasDataList=[]
        self.angleSetup=0
        self.configTable=0
        self.antennaType=0
        self.chirp=0
        self.nofAngleValues=0
        self.angle=0
        self.errorStatus=0
        self.avgNormLevelDb=0
        self.dataTableString=''
        self.setup(angleSetupData,configTable,chirp,parseDataList,ptu2InputConf)

    # Set string.
    def setStr(self):

        if self.angleSetup.diagType =='NoDiagType':
            diagramTypeStr=''
        else:
            diagramTypeStr='_'+self.angleSetup.diagType

        self.dataTableString=self.ptu2InputConf.getStr()+str(diagramTypeStr)+' - AntType-'+str(self.angleSetup.antennaType)+'_NofPts='+str(self.angleSetup.nofAnglePoints)+'_StepSize='+str(self.angleSetup.angleStepSize)\
        +'_RefPt='+str(self.angleSetup.refAnglePoint)+' - Chirp'+str(self.chirp)+' - Tab'+str(self.configTable.antennaPatternIdx)

    # Get string.
    def getStr(self):
        self.setStr()
        return self.dataTableString

    # Set angle data
    def setAngle(self):
        if self.errorStatus==0:
            self.angle = calc.getPtuAngVals(self.tableMeasDataList[0].parseData)
            self.nofAngleValues=len(self.angle)
            if self.ptu2InputConf.targetAngle == 1:
                self.angle = calc.getTarAngVals(self.tableMeasDataList[0].parseData, getUpsideDown(self.tableMeasDataList[0].parseData,self.ptu2InputConf) )
                # self.angle=mapInputOutput(self.angle,
                #                self.angleSetup.nofAnglePoints,
                #                self.angleSetup.angleStepSize,
                #                self.angleSetup.refAnglePoint,MAX_ANGLE_POINTS, setup["targetAngle"], setup["upsideDown"])

        else:
            myLogger.debug("Angle data not available. Empty parsed data list: AntPatTable" +str(self.configTable.antennaPatternIdx)+' - '+ self.configTable.getStr())

    # Set the list of data for certain Antenna Type, Tx antennas and Rotation axis from the Configured tables, Angle setup, and Chirp.
    def setTableMeasDataList(self,parseDataList):

        upsideDownData = getUpsideDownData(parseDataList, self.ptu2InputConf.upsideDownInfo)

        if len(upsideDownData) != 0:

            if self.ptu2InputConf.diagTypeInfo==1:

                myLogger.debug('Getting PTU2 parsed data WITH dagram type info, AntPatTable' +str(self.configTable.antennaPatternIdx)+': '+ str(self.configTable.getStr()))

                applicableTxAntData = getApplicableTxAntIndData(self.configTable.applTxAntInd, self.configTable.nofApplicableTxAnt, upsideDownData)
                txAntDiagTypeCombData=[]

                if len(applicableTxAntData)==0:
                    myLogger.info('No PTU2 files with Tx Antenna info found. Looking for matching PTU2 without Tx Antenna info.')

                    noTxAntData = getNoTxAntData(upsideDownData)
                    if len(noTxAntData)!=0:
                        txAntDiagTypeCombData=getDesiredDiagTypeData(antennaDiagramTypeMapping(self.configTable.rotationAxis), noTxAntData)
                        myLogger.info('PTU2 files without Tx Antenna info found. Processng as configured Tx Antenna.')
                    else:
                        myLogger.error('No PTU2 files without Tx Antenna info found.')
                        self.errorStatus=1
                else:

                    txAntDiagTypeCombData = getDesiredDiagTypeData(antennaDiagramTypeMapping(self.configTable.rotationAxis), applicableTxAntData)

                if len(txAntDiagTypeCombData)!=0:

                    for txAntDiagTypeIdx in range(len(txAntDiagTypeCombData)):

                        fileNamePath=txAntDiagTypeCombData[txAntDiagTypeIdx].fileNamePath
                        myLogger.debug('LISTED: ' +fileNamePath )

                        nofAngPts=getNofAnglePoints(txAntDiagTypeCombData[txAntDiagTypeIdx])
                        nofAngPtsSelf=self.angleSetup.nofAnglePoints
                        # myLogger.debug('NofAngPts: ' + str(nofAngPts)+', NofAngPtsSelf: '+str(nofAngPtsSelf))

                        angleStep=getAngleStep(txAntDiagTypeCombData[txAntDiagTypeIdx])
                        angleStepSelf=self.angleSetup.angleStepSize
                        # myLogger.debug('AngleStep: ' + str(nofAngPts) + ', AngleStepSelf: ' + str(nofAngPtsSelf))

                        refAngPt=getRefAnglePoint(txAntDiagTypeCombData[txAntDiagTypeIdx])
                        refAngPtSelf=self.angleSetup.refAnglePoint
                        # myLogger.debug('RefAngPt: ' + str(nofAngPts) + ', RefAngPtSelf: ' + str(nofAngPtsSelf))

                        antennaType = txAntDiagTypeCombData[txAntDiagTypeIdx].antennaInfo.antennaType
                        antennaTypeSelf = self.angleSetup.antennaType

                        chirpActive=txAntDiagTypeCombData[txAntDiagTypeIdx].chirpInfoList[self.chirp].active
                        # myLogger.debug('ChirpSelf: ' + str(self.chirp)+', Active: '+str(chirpActive))

                        if (antennaType==antennaTypeSelf) and (nofAngPts ==nofAngPtsSelf ) and (angleStep == angleStepSelf) and (abs(refAngPt - refAngPtSelf)<0.1) and (chirpActive==1):
                            tableMeasData=TableMeasData()
                            tableMeasData.parseData=txAntDiagTypeCombData[txAntDiagTypeIdx]
                            self.tableMeasDataList.append(tableMeasData)
                            myLogger.info('PTU2 file added: ' + tableMeasData.parseData.fileNamePath)


                    if len(self.tableMeasDataList)==0:
                        self.errorStatus=1
                        myLogger.error('No PTU2 parsed data files WITH diagram type info found: Table'+str(self.configTable.antennaPatternIdx)+' - '+ self.configTable.getStr())
                    else:
                        maxNofAvailableTxAnt = getNofAvailableTxAnt(self.tableMeasDataList)
                        if self.configTable.nofApplicableTxAnt > maxNofAvailableTxAnt:
                            myLogger.error('Number of applicable Tx Antenna indices: '+str(self.configTable.nofApplicableTxAnt)+ ', surpasses the total number of available Tx Antenna: '+str(maxNofAvailableTxAnt)+'.')
                            self.errorStatus=1
                        else:
                            self.errorStatus=0
                else:
                    self.errorStatus=1
                    myLogger.error("No data for Diagram Type found: "+antennaDiagramTypeMapping(self.configTable.rotationAxis)+'.')
            else:

                myLogger.debug('Getting PTU2 parsed data WITHOUT dagram type info, Table' + str(self.configTable.antennaPatternIdx) + ': ' + str(self.configTable.getStr()))

                applicableTxAntData = getApplicableTxAntIndData(self.configTable.applTxAntInd,self.configTable.nofApplicableTxAnt, upsideDownData)

                if len(applicableTxAntData) == 0:
                    myLogger.info('No PTU2 files with Tx Antenna info found. Looking for matching PTU2 without Tx Antenna info.')

                    noTxAntData = getNoTxAntData(upsideDownData)
                    noDiagTypeList=[]

                    if len(noTxAntData) != 0:
                        noDiagTypeList = getNonDiagTypeFilesList(noTxAntData)
                        myLogger.info('PTU2 files without Tx Antenna info found. Processng as configured Tx Antenna.')
                    else:
                        myLogger.error('No PTU2 files without Tx Antenna info found.')
                        self.errorStatus=1
                else:

                    noDiagTypeList = getNonDiagTypeFilesList(applicableTxAntData)

                if len(noDiagTypeList) != 0:

                    for txAntDiagTypeIdx in range(len(noDiagTypeList)):
                        if getNofAnglePoints(noDiagTypeList[txAntDiagTypeIdx]) == self.angleSetup.nofAnglePoints and \
                                getAngleStep(noDiagTypeList[txAntDiagTypeIdx]) == self.angleSetup.angleStepSize and \
                                abs (getRefAnglePoint(noDiagTypeList[txAntDiagTypeIdx]) - self.angleSetup.refAnglePoint)<0.1 \
                                and noDiagTypeList[txAntDiagTypeIdx].antennaInfo.antennaType==self.angleSetup.antennaType \
                                and noDiagTypeList[txAntDiagTypeIdx].chirpInfoList[self.chirp].active == 1:
                            tableMeasData = TableMeasData()
                            tableMeasData.parseData = noDiagTypeList[txAntDiagTypeIdx]
                            self.tableMeasDataList.append(tableMeasData)
                            myLogger.info('PTU2 file added: ' + tableMeasData.parseData.fileNamePath)

                    if len(self.tableMeasDataList)==0:
                        self.errorStatus=1
                        myLogger.error('No PTU2 parsed data files WITHOUT diagram type info found: Table'+str(self.configTable.antennaPatternIdx)+' - '+ self.configTable.getStr())
                    else:
                        maxNofAvailableTxAnt = getNofAvailableTxAnt(self.tableMeasDataList)
                        if self.configTable.nofApplicableTxAnt > maxNofAvailableTxAnt:
                            myLogger.error('Number of applicable Tx Antenna indices: '+str(self.configTable.nofApplicableTxAnt)+ ', surpasses the total number of available Tx Antenna: '+str(maxNofAvailableTxAnt)+'.')
                            self.errorStatus=1
                        else:
                            self.errorStatus=0

                else:
                    self.errorStatus = 1
                    myLogger.error("No data for WITHOUT diagram type info found: " + antennaDiagramTypeMapping(self.configTable.rotationAxis) + '.')
        else:
            self.errorStatus = 1
            myLogger.error("No PTU2 parsed data files found, upsideDownInfo: " + str(self.ptu2InputConf.upsideDownInfo) + '.')


    # Set the Measuremenr Data list with normalised level values.
    def setTableMeasDataVals(self):
        if self.errorStatus==0:

            for tmDataIdx in range(len(self.tableMeasDataList)):

                self.tableMeasDataList[tmDataIdx].normLevelList=self.setNormLevelData(self.configTable, self.chirp, self.tableMeasDataList[tmDataIdx].parseData)
                #
                # for rxAnt in range(len(self.tableMeasDataList[tmDataIdx].normLevelList)):
                #
                #     self.tableMeasDataList[tmDataIdx].normLevelList[rxAnt] = mapInputOutput(self.tableMeasDataList[tmDataIdx].normLevelList[rxAnt],
                #                                           self.angleSetup.nofAnglePoints,
                #                                             self.angleSetup.angleStepSize,
                #                                         self.angleSetup.refAnglePoint,MAX_ANGLE_POINTS, setup["targetAngle"], setup["upsideDown"])
                self.setRxAntAvgNormLevel(self.tableMeasDataList[tmDataIdx])

    # Set normalised level data for the data table input parameters and a certain PTU2 parsed file from the Measurement Data list.
    def setNormLevelData(self,configTable,chirp,parseData):
        normLevelList = []
        if configTable.nofApplicableRxAnt <= parseData.antennaInfo.nofAntennas:
            for rxAntIdx in range(configTable.nofApplicableRxAnt):
                normLevel = calc.getNormAntLevel(configTable.applRxAntInd[rxAntIdx], chirp, parseData)
                normLevelList.append(normLevel)
        else:
            self.errorStatus=1
            myLogger.error("The number of the applicable Rx antennas surpasses the number of sensor antennas: Table"+str(configTable.antennaPatternIdx)+' - '+configTable.getStr()+' - PTU2 file - '+parseData.fileNamePath)
        return normLevelList

    # Set the average normalised level of the configured Rx antennas for a certain PTU2 parsed file from the Meas Data list.
    def setRxAntAvgNormLevel(self,tableMeasData):
        if self.errorStatus==0:
            normLevelSum=0
            for rxAnt in range(len(tableMeasData.normLevelList)):
                normLevelW = 10 ** (tableMeasData.normLevelList[rxAnt] / 10)
                normLevelSum += normLevelW
            avgNormLevel=normLevelSum/len(tableMeasData.normLevelList)
            tableMeasData.rxAntAvgNormLevelW=avgNormLevel
            tableMeasData.rxAntAvgNormLevelDb =10 * np.log10(tableMeasData.rxAntAvgNormLevelW)


    # Set the values for given input parameters.
    def setup(self,angleSetupData,configTable,chirp,parseDataList,ptu2InputConf):
        self.ptu2InputConf=ptu2InputConf
        # self.antennaType=antennaType
        self.angleSetup=angleSetupData
        self.configTable=configTable
        self.chirp=chirp
        self.avgNormLevelDb=0
        self.setTableMeasDataList(parseDataList)
        self.setTableMeasDataVals()
        self.setAngle()

# Class containing the data for Antenna Pattern WITHOUT diagram type info, for certain Configured Table and certain Antenna Type, Angle setup, and Chirp.
# class DataTableNoDiagType(DataTable):
#     # Set the list of data for No Diagram Type info setup, certain Antenna Type, Tx antennas, Configured tables, Angle setup, and Chirp.
#     def setTableMeasDataList(self, parseDataList):
#         myLogger.debug('Getting PTU2 parsed data WITHOUT dagram type info, Table' + str(self.configTable.antennaPatternIdx) + ': ' + str(self.configTable.getStr()))
#
#         antennaTypeData = getAntennaTypeData(self.antennaType, parseDataList)
#
#         if len(antennaTypeData) != 0:
#
#             applicableTxAntData = getApplicableTxAntIndData(self.configTable.applTxAntInd,self.configTable.nofApplicableTxAnt, upsideDownData)
#
#             if len(applicableTxAntData)==0:
#                 myLogger.debug('No applicable Tx Antenna data found.')
#                 noTxAntData = getNoTxAntData(antennaTypeData)
#
#                 noDiagTypeList = getNonDiagTypeFilesList(noTxAntData)
#             else:
#
#                 noDiagTypeList = getNonDiagTypeFilesList(applicableTxAntData)
#
#             if len(noDiagTypeList)!=0:
#
#                 for txAntDiagTypeIdx in range(len(noDiagTypeList)):
#                     if getNofAnglePoints(noDiagTypeList[txAntDiagTypeIdx], self.ptu2InputConf.targetAngle,
#                                          self.ptu2InputConf.upsideDown) == self.angleSetup.nofAnglePoints and \
#                             getAngleStep(noDiagTypeList[txAntDiagTypeIdx], self.ptu2InputConf.targetAngle,
#                                          self.ptu2InputConf.upsideDown) == self.angleSetup.angleStepSize and \
#                             getRefAnglePoint(noDiagTypeList[txAntDiagTypeIdx], self.ptu2InputConf.targetAngle,
#                                              self.ptu2InputConf.upsideDown) == self.angleSetup.refAnglePoint \
#                             and noDiagTypeList[txAntDiagTypeIdx].chirpInfoList[self.chirp].active==1:
#                         tableMeasData = TableMeasData()
#                         tableMeasData.parseData = noDiagTypeList[txAntDiagTypeIdx]
#                         self.tableMeasDataList.append(tableMeasData)
#                         myLogger.info('PTU2 file added: ' + tableMeasData.parseData.fileNamePath)
#
#                 if len(self.tableMeasDataList) == 0:
#                     self.errorStatus = 1
#                     myLogger.debug('No PTU2 parsed data files WITHOUT diagram type info found: Table' + str(
#                         self.configTable.antennaPatternIdx) + '- ' + self.configTable.getStr())
#                 else:
#                     self.errorStatus = 0
#             else:
#                 self.errorStatus = 1
#                 myLogger.debug("No data for WITHOUT diagram type info found: " + antennaDiagramTypeMapping(self.configTable.rotationAxis) + '.')
#
#         else:
#             self.errorStatus=1
#             myLogger.debug("No PTU2 parsed data files WITHOUT diagram type info found: Antenna Type: " +str(self.antennaType)+'.')

# Class containing the norm level data and PTU2 parsed data for certain element of the list containing data for certain input.
class TableMeasData():
    def __init__(self):
        self.normLevelList = []
        self.rxAntAvgNormLevelW=0
        self.rxAntAvgNormLevelDb = 0
        self.parseData=0

# Class containing the parameters for angle setup.
class AngleSetup():
    def __init__(self,angleSetupData):
        self.antennaType=0
        self.diagType = ''
        self.nofAnglePoints=0
        self.angleStepSize=0
        self.refAnglePoint=0
        self.angleSetupStr = ''
        self.setAngle(angleSetupData)
        self.setStr()

    # Set string.
    def setStr(self):
        self.angleSetupStr='AntType-'+str(self.antennaType)+'_'+str(self.diagType)+'_NofPts='+str(self.nofAnglePoints)+'_StepSize='+str(self.angleStepSize)+'_RefPt='+str(self.refAnglePoint)

    # Set values according to given input parameters.
    def setAngle(self,angleSetupData):
        self.antennaType=angleSetupData["antType"]
        self.diagType = angleSetupData["diagType"]
        self.nofAnglePoints = angleSetupData["nofAnglePoints"]
        self.angleStepSize =angleSetupData["angleStepSize"]
        self.refAnglePoint = angleSetupData["refAnglePoint"]

# Class containing the prarameters of a configured table.
class ConfigTable():
    def __init__(self,configDict):
        self.antennaPatternIdx=0
        self.patternType=0
        self.rotationAxis=0
        self.nofApplicableTxAnt=0
        self.nofApplicableRxAnt=0
        self.applTxAntInd=[]
        self.applRxAntInd=[]
        self.gainCalcMethod=0
        self.errorStatus=0
        self.errorStringList=[]
        self.rxErrorStringList = []
        self.txErrorStringList = []
        self.confTabStr=''
        self.setTable(configDict)

        # Check configuration errors for Tx antennas.
        if self.checkAppAntConf(self.nofApplicableTxAnt,self.applTxAntInd)==1:
            self.txErrorStringList=self.errorStringList
            myLogger.error("Invalid configuration for applicable Tx antennas: Table"+str(self.antennaPatternIdx)+' - '+self.getStr())
            for i in range(len(self.txErrorStringList)):
                errorString=self.txErrorStringList[i]
                myLogger.error(errorString)

        # Check configuration errors for Rx antennas.
        if self.checkAppAntConf(self.nofApplicableRxAnt, self.applRxAntInd)==1:
            self.rxErrorStringList = self.errorStringList
            myLogger.error("Invalid configuration for applicable Rx antennas: Table"+str(self.antennaPatternIdx)+' - '+self.getStr())
            for i in range(len(self.rxErrorStringList)):
                errorString=self.rxErrorStringList[i]
                myLogger.error(errorString)

        # Check congiguration for pattern type, and rotation axis.
        self.checkPatternType()
        self.checkRotationAxis()

    # Check congiguration for rotation axis.
    def checkRotationAxis(self):

        allowedVals = [0, 1, 2, 3]

        allowedValsStr = '0, 1 ,2, 3'

        if self.rotationAxis not in allowedVals:
            self.errorStatus = 1
            myLogger.error("Invalid rotation axis, Table" + str(self.antennaPatternIdx) + ' - ' + self.getStr() + '. It should be one of these values: ' + allowedValsStr + '.')

    # Check congiguration for pattern type.
    def checkPatternType(self):

        allowedVals=[0,1,2,3]

        allowedValsStr = '0, 1 ,2, 3'

        if self.patternType not in allowedVals:
            self.errorStatus=1
            myLogger.error("Invalid pattern type, Table"+str(self.antennaPatternIdx)+' - '+self.getStr()+'. It should be one of these values: '+allowedValsStr+'.')

    # Check applicable antenas configuration.
    def checkAppAntConf(self,nofAppAnt,appAntIndList):
        ret=0
        errorString=''
        checkNoInc=0
        checkNullAppAnt=0
        checkMaxNumAppAnt=0
        checkAppAntIndListLen=0
        checkNoMatch=0
        errorStringList=[]

        # if nofAppAnt>MAX_NUM_ANT:
        #     checkMaxNumAppAnt=1
        #     errorString='The number of applicable antennas is larger than the predefined maximum: ' + str(MAX_NUM_ANT)+'.'
        #     errorStringList.append(errorString)

        if nofAppAnt<=0:
            checkNullAppAnt=1
            errorString='The number of applicable antennas is zero or less.'
            errorStringList.append(errorString)

        # if len(appAntIndList)>MAX_NUM_ANT:
        #     checkAppAntIndListLen=1
        #     errorString='The length of the applicable antenna incices list is larger than the predefined maximal number of applicable antennas: ' + str(MAX_NUM_ANT)+'.'
        #     errorStringList.append(errorString)

        # nofNullInd=len(appAntIndList)-nofAppAnt
        # count=0
        # if nofAppAnt==1:
        #     stop=0
        # else:
        #     stop=-1
        # for i in range(len(appAntIndList)-1,stop,-1):
        #     if appAntIndList[i]==0:
        #         count=count+1
        #     else:
        #         break
        # if count!=nofNullInd:
        #     checkNoMatch=1
        #     errorString='Applicable antenna indices values do not match the number of applicable antennas.'
        #     errorStringList.append(errorString)

        if nofAppAnt>len(appAntIndList):
            stop=len(appAntIndList)
        else:
            stop=nofAppAnt
        temp = -1
        for i in range(stop):
            if appAntIndList[i] > temp:
                temp = appAntIndList[i]
            else:
                checkNoInc = 1
        if checkNoInc ==1:
            errorString='Applicable antenna indices are not in incremental order.'
            errorStringList.append(errorString)

        if len(appAntIndList)!=nofAppAnt:
            checkNoMatch=1
            errorString='Applicable antenna indices values do not match the number of applicable antennas.'
            errorStringList.append(errorString)

        self.errorStringList=errorStringList

        if checkNoInc==1 or checkNullAppAnt==1 or checkNoMatch==1:
            self.errorStatus=1
            ret=1
        return ret

    # Set string
    def setStr(self):
        confTableDict=dict()

        confTableDict.update({"antenna_pattern_idx":self.antennaPatternIdx,"patternType":self.patternType,"rotationAxis":self.rotationAxis,
                     "nofApplicableTxAnt":self.nofApplicableTxAnt,"nofApplicableRxAnt":self.nofApplicableRxAnt,
                     "applTxAntInd":self.applTxAntInd,"applRxAntInd":self.applRxAntInd,"gainCalcMethod":self.gainCalcMethod})

        self.confTabStr=str(confTableDict)

    # Get string
    def getStr(self):
        self.setStr()
        return self.confTabStr

    # Set values according to certain input.
    def setTable(self,configDict):
        self.antennaPatternIdx=configDict["antenna_pattern_idx"]
        self.patternType=configDict["patternType"]
        self.rotationAxis=configDict["rotationAxis"]
        self.nofApplicableTxAnt=configDict["nofApplicableTxAnt"]
        self.nofApplicableRxAnt=configDict["nofApplicableRxAnt"]
        self.applTxAntInd=configDict["applTxAntInd"]
        self.applRxAntInd=configDict["applRxAntInd"]
        self.gainCalcMethod=configDict["gainCalcMethod"]
        self.setStr()

# # Mapping string for the antenna diagram types
# def getAntDiagType(inputString):
#     if inputString == 'NVM_ANTPAT_ROT_AXIS_Z' or inputString ==1:
#         antennaDiagType = 'Az'
#     else:
#         antennaDiagType = 'El'
#     return antennaDiagType

# Get the PTU2 parsed data for desired diagram type.
def getDesiredDiagTypeData(diagramType, parseDataList):
    myLogger.debug("Looking for " +str(diagramType) +' PTU2 parsed data.')
    outputList = []
    for pdIdx in range(len(parseDataList)):
        if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound == 1:
            if parseDataList[pdIdx].measSetupInfo.antennaDiagramType == diagramType:
                outputData = parseDataList[pdIdx]
                outputList.append(outputData)

    # myLogger.debug('outputList#########')
    # for i in range(len(outputList)):
        # myLogger.debug('PTU#####outputLIST*******: ' + outputList[i].fileNamePath)
        # myLogger.debug('Chirp0: ' + str(outputList[i].chirpInfoList[0].active))
        # myLogger.debug('Chirp1: ' + str(outputList[i].chirpInfoList[1].active))
        # myLogger.debug('Chirp2: ' + str(outputList[i].chirpInfoList[2].active))
        # myLogger.debug('Chirp3: ' + str(outputList[i].chirpInfoList[3].active))
    return outputList

# Get antenna types list.
def getAntennaTypeData(antennaType,parseData):
    outputList=[]
    myLogger.debug("Looking for Antenna Type - " +str(antennaType) +' PTU2 parsed data.')
    for pdIdx in range(len(parseData)):
        if parseData[pdIdx].antennaInfo.antennaType==antennaType:
            outputList.append(parseData[pdIdx])
            myLogger.debug('File added: '+parseData[pdIdx].fileNamePath+'.')

    return outputList
# Get list of the parsed data with upsideDown info.
def getUpsideDownData(parseDataList,upsideDownInfo):
    outputList=[]
    if upsideDownInfo ==1:
        for pdIdx in range(len(parseDataList)):
            if parseDataList[pdIdx].measSetupInfo.sensorUpsideDownFound==1:
                outputList.append(parseDataList[pdIdx])
    else:
        for pdIdx in range(len(parseDataList)):
            if parseDataList[pdIdx].measSetupInfo.sensorUpsideDownFound!=1:
                outputList.append(parseDataList[pdIdx])
    return outputList

def getUpsideDown(parseData,ptu2InputConfig):
    if ptu2InputConfig.upsideDownInfo==0:
        upsideDown = ptu2InputConfig.upsideDown
    else:
        upsideDown = parseData.measSetupInfo.sensorUpsideDown
    return upsideDown

# Get PTU2 parsed data for PTU2 files without Tx antenna info.
def getNoTxAntData(parseDataList):
    outputList=[]
    myLogger.debug('Getting PTU2 parsed data without Tx antenna info.')
    for i in range(len(parseDataList)):
        if parseDataList[i].fileTitle.txAntFound==0:
            outputList.append(parseDataList[i])
            myLogger.debug('File added: '+parseDataList[i].fileNamePath+'.')

    return outputList

# Get PTU2 parsed data for PTU2 files without Tx antenna info.
def getDesiredTxAntFilesList(txAnt, parseDataList):
    outputList = []
    for pdIdx in range(len(parseDataList)):
        if parseDataList[pdIdx].fileTitle.txAntFound == 1 and parseDataList[pdIdx].fileTitle.txAnt == txAnt:
            actuaData = parseDataList[pdIdx]
            outputList.append(actuaData)
            # myLogger.debug('PTU#####: ' + parseDataList[pdIdx].fileNamePath)
            # myLogger.debug('Chirp0: '+str(parseDataList[pdIdx].chirpInfoList[0].active))
            # myLogger.debug('Chirp1: ' + str(parseDataList[pdIdx].chirpInfoList[1].active))
            # myLogger.debug('Chirp2: ' + str(parseDataList[pdIdx].chirpInfoList[2].active))
            # myLogger.debug('Chirp3: ' + str(parseDataList[pdIdx].chirpInfoList[3].active))

    return outputList

# Get the list of the lists containing PTU2 parsed data for given Tx antennas, according to applicable Tx antenna indices.
def getApplicableTxAntIndData(appTxAntIndArray, nofApplicableTxAnt, parseDataList):
    dataList = []
    arrayLen = len(appTxAntIndArray)
    if nofApplicableTxAnt <= arrayLen:
        for txAntInd in range(nofApplicableTxAnt):
            data = getDesiredTxAntFilesList(appTxAntIndArray[txAntInd], parseDataList)
            dataList += data
            myLogger.debug("Looking for TxAnt"+str(appTxAntIndArray[txAntInd]) +' PTU2 parsed data.')

    # myLogger.debug('DataList#########')
    # for i in range(len(dataList)):
        # myLogger.debug('PTU#####DATALIST: ' + dataList[i].fileNamePath)
        # myLogger.debug('Chirp0: ' + str(dataList[i].chirpInfoList[0].active))
        # myLogger.debug('Chirp1: ' + str(dataList[i].chirpInfoList[1].active))
        # myLogger.debug('Chirp2: ' + str(dataList[i].chirpInfoList[2].active))
        # myLogger.debug('Chirp3: ' + str(dataList[i].chirpInfoList[3].active))

    return dataList
# String mapping for diagram type.
def antennaDiagramTypeMapping(rotationAxis):
    if rotationAxis == 2:
        antDiagType = 'El'
    else:
        antDiagType = 'Az'
    return antDiagType


# Get the num. of diagrams from the parsed data
def getNofDiagrams(parseDataList):
    diagramIdxList=[]
    for pdIdx in range(len(parseDataList)):
        diagram=parseDataList[pdIdx].fileTitle.diagram
        diagramIdxList.append(diagram)
    if len(diagramIdxList)!=0:
        nofDiagrams=max(diagramIdxList)+1
    else:
        nofDiagrams=0
    return nofDiagrams

# Get the list of all sensors from the parsed data
def getSensorList(parseDataList):
    sensor = ''
    sensorList = []
    for listIdx in range(len(parseDataList)):
        if parseDataList[listIdx].fileTitle.sensorFound==1:
            if parseDataList[listIdx].fileTitle.sensor!= sensor:
                sensor = parseDataList[listIdx].fileTitle.sensor
                sensorList.append(parseDataList[listIdx].fileTitle.sensor)
    return sensorList

# Get the list of all antenna types from the parsed data
def getAntennaTypesList(parseDataList,ptu2InputConf):
    antennaType = 0
    antennaTypeList = []
    if ptu2InputConf.diagTypeInfo == 0:
        for listIdx in range(len(parseDataList)):
            if parseDataList[listIdx].measSetupInfo.antennaDiagramTypeFound != 1:
                if parseDataList[listIdx].antennaInfo.antennaType!= antennaType:
                    antennaType = parseDataList[listIdx].antennaInfo.antennaType
                    antennaTypeList.append(parseDataList[listIdx].antennaInfo.antennaType)
    else:
        for listIdx in range(len(parseDataList)):
            if parseDataList[listIdx].measSetupInfo.antennaDiagramTypeFound == 1:
                if parseDataList[listIdx].antennaInfo.antennaType!= antennaType:
                    antennaType = parseDataList[listIdx].antennaInfo.antennaType
                    antennaTypeList.append(parseDataList[listIdx].antennaInfo.antennaType)

    antTypeSet=set(antennaTypeList)
    outputList=list(antTypeSet)

    return outputList


# Get number of TxAnt per sensor
def getNofTxAntPerSensor(sensor, parseDataList):
    actualTxAnt=0
    txAntList=[]

    for pdIdx in range(len(parseDataList)):
        if parseDataList[pdIdx].antennaInfo.fileTitle.sensorFound==1:
            if sensor == parseDataList[pdIdx].fileTitle.sensor:
                if parseDataList[pdIdx].antennaInfo.fileTitle.txAntFound == 1:
                    actualTxAnt = parseDataList[pdIdx].fileTitle.txAnt
                    txAntList.append(actualTxAnt)
    return max(txAntList)

# Get PTU2 parsed data for files without diagram type info.
def getNonDiagTypeFilesList(parseDataList):
    outputList = []
    for pdIdx in range(len(parseDataList)):
        if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound != 1:
                outputData = parseDataList[pdIdx]
                outputList.append(outputData)
    return outputList

# Get angle step from PTU2 parsed data, for angle setup configuration.
def getAngleStep(parseData):
    # if isTargetAngle ==1:
    #     angle=calc.getTarAngVals(parseData,upsideDown)
    # else:
    #     angle=calc.getPtuAngVals(parseData)
    angle = calc.getPtuAngVals(parseData)
    medianList=[]
    for i in range(1,5):
        firstPoint=round(angle[i-1],1)
        lastPoint=round(angle[i],1)
        step=abs(firstPoint-lastPoint)
        medianList.append(step)
    median=np.median(medianList)
    return median

# Get number fo angle points from PTU2 parsed data, for angle setup configuration.
def getNofAnglePoints(parseData):
    # if isTargetAngle == 1:
    #     length=len(calc.getTarAngVals(parseData, upsideDown))
    # else:
    #     length=len(calc.getPtuAngVals(parseData))
    length=len(calc.getPtuAngVals(parseData))
    return length

# Get the reference angle point(the first point from the angle array) from PTU2 parsed data, for angle setup configuration.
def getRefAnglePoint(parseData):
    # if isTargetAngle ==1:
    #     angle=calc.getTarAngVals(parseData,upsideDown)
    # else:
    #     angle=calc.getPtuAngVals(parseData)
    angle = calc.getPtuAngVals(parseData)
    refAnglePoint=round(angle[0],1)
    return refAnglePoint

# Get a list of the appearing angle setup (nof. angle points, angle step size, ref. angle point) and diagram type info (respectively Non diagram type info) combintains.
def getAppearingAngleSetupComb(parseDataList):

    combDictList=[]
    outputList=[]
    myLogger.info('Getting angle setup list:')


    for pdIdx in range(len(parseDataList)):

        combDict = dict()
        # step=getAngleStep(parseDataList[pdIdx],isTargetAngle,upsideDown)
        # nofAnglePoints=getNofAnglePoints(parseDataList[pdIdx],isTargetAngle,upsideDown)
        # refAnglePoint=getRefAnglePoint(parseDataList[pdIdx],isTargetAngle,upsideDown)

        step=getAngleStep(parseDataList[pdIdx])
        nofAnglePoints=getNofAnglePoints(parseDataList[pdIdx])
        refAnglePoint=getRefAnglePoint(parseDataList[pdIdx])

        antennaType=parseDataList[pdIdx].antennaInfo.antennaType

        if parseDataList[pdIdx].measSetupInfo.antennaDiagramTypeFound==1:
            diagType=parseDataList[pdIdx].measSetupInfo.antennaDiagramType
        else:
            diagType ='NoDiagType'

        combDict.update({"antType":antennaType,"diagType":diagType,"nofAnglePoints": nofAnglePoints, "angleStepSize": step,"refAnglePoint":refAnglePoint})
        combDictList.append(combDict)

    # outputList=list(np.unique(np.array(combDictList)))

    for pdIdx in range(len(combDictList)):
        if pdIdx==0:
            outputList.append(combDictList[pdIdx])
            myLogger.debug('Angle setup comb. added: '+ 'AntType-'+str(combDictList[pdIdx]["antType"]) + '_DiagType='+combDictList[pdIdx]["diagType"]+ '_NofAngPts=' + str(combDictList[pdIdx]["nofAnglePoints"]) + '_AngStepSize=' + \
                          str(combDictList[pdIdx]["angleStepSize"]) + '_RefAngPt=' + str(combDictList[pdIdx]["refAnglePoint"]))
        isDifferentList=[]
        for outputIdx in range(len(outputList)):
            isDifferent=0
            if outputList[outputIdx]["antType"]!=combDictList[pdIdx]["antType"]\
                    or outputList[outputIdx]["diagType"]!=combDictList[pdIdx]["diagType"]\
                    or outputList[outputIdx]["nofAnglePoints"]!=combDictList[pdIdx]["nofAnglePoints"]\
                    or outputList[outputIdx]["angleStepSize"]!=combDictList[pdIdx]["angleStepSize"]\
                    or abs(outputList[outputIdx]["refAnglePoint"]- combDictList[pdIdx]["refAnglePoint"])>=0.1:
               isDifferent=1
            isDifferentList.append(isDifferent)

        found=1
        for difIdx in range(len(isDifferentList)):
            if isDifferentList[difIdx] ==0:
                found=0
                break

        if found ==1:
            outputList.append(combDictList[pdIdx])
            myLogger.debug('Angle setup comb. added: '+ 'AntType-'+str(combDictList[pdIdx]["antType"]) + '_DiagType='+combDictList[pdIdx]["diagType"]+  '_NofAngPts=' + str(combDictList[pdIdx]["nofAnglePoints"]) + '_AngStepSize=' + str(combDictList[pdIdx]["angleStepSize"]) + '_RefAngPt=' + str(combDictList[pdIdx]["refAnglePoint"])+'.')
                # outputList.append(combDictList[pdIdx])

    return outputList

# Map the input measurement data array to the maximal number of allowed angle points.
def mapInputOutput(inputArr,nofAngPts, stepSize, refPoint,maxAnglePoints,isTargetAngle,upsideDown):
    if isTargetAngle==1 and upsideDown==0:
        stepSize=-stepSize
    firstPoint=refPoint
    lastPoint=refPoint +((nofAngPts-1)*stepSize)
    valuesRange = abs(firstPoint - lastPoint)
    nofSlices=nofAngPts-1

    # outputDict=dict()
    if nofAngPts > maxAnglePoints:

        n = 1
        outputStepSize=n*stepSize
        outputNofSlices=int(valuesRange / abs(outputStepSize))
        outputAngPts = 1+ outputNofSlices
        outputLastPoint = refPoint + ((outputAngPts - 1) * outputStepSize)
        # div=nofSlices/outputNofSlices

        while outputAngPts>91:
            n = n + 1
            outputStepSize = n * stepSize
            outputNofSlices = int(valuesRange / abs(outputStepSize))
            outputAngPts = 1 + outputNofSlices
            outputLastPoint = refPoint + ((outputAngPts - 1) * outputStepSize)
            # div = nofSlices / outputNofSlices

        outputArray=inputArr[0::n]
        myLogger.debug("Angle points reduced to: MAX_ANGLE_POINTS: " + str(maxAnglePoints) + ", StepSize: "+str(n)+'.')
        # outputDataArr=inputDataArr[0::n]

        # outputDict.update({"angleArray":outputArray,"normLevelArray":outputDataArr})
    else:
        outputArray = inputArr
    return outputArray

# Get the configured chirps
def getConfChirpList(chirpSetup):
    chirpList = []
    for chirp, set in chirpSetup.items():
        if set == 1:
            parts=chirp.split('chirp')
            chirpNum=parts[1]
            chirpNum=int(chirpNum)
            chirpList.append(chirpNum)
    return chirpList

# Get the actual number of Rx channels, per specific graph and a PTU2 file.
def getNofActualRxChannels(graphList, graphIdx, parseDataList, parseDataIdx):
    if graphList[graphIdx].endswith('evel'):
        actualNofRxCh = parseDataList[parseDataIdx].antennaInfo.nofAntennas
    else:
        actualNofRxCh = parseDataList[parseDataIdx].antennaInfo.nofAntennaDiffValues
    return actualNofRxCh

# Get the actual number of Rx channels, per specific graph and a PTU2 file.
def getNofActualRxChannels(graph, parseDataList, pdIdx):
    if graph.endswith('evel'):
        actualNofRxCh = parseDataList[pdIdx].antennaInfo.nofAntennas
    else:
        actualNofRxCh = parseDataList[pdIdx].antennaInfo.nofAntennaDiffValues
    return actualNofRxCh

# Get the maximal number of Rx channels for specific graph, of all PTU2 files.
def getMaxNofRxCh(graph,parseDataList):
    nofRxChList = []
    for pdIdx in range(len(parseDataList)):
        nofRxCh = getNofActualRxChannels(graph, parseDataList, pdIdx)
        nofRxChList.append(nofRxCh)
    return max(nofRxChList)

# Get diagram type info string
def getDiagTypeInfoString(ptu2InputConf):
    if ptu2InputConf['diagTypeInfo']==1:
        ret='WITH diagram type info'
    else:
        ret ='WITHOUT diagram type info'
    return ret

def getNofAvailableTxAnt(tableMeasDataList):
    txAntList=[]
    for pdIdx in range(len(tableMeasDataList)):
        if tableMeasDataList[pdIdx].parseData.fileTitle.txAntFound == 1:
            actualTxAnt =tableMeasDataList[pdIdx].parseData.fileTitle.txAnt
            txAntList.append(actualTxAnt)
    txAntSet=set(txAntList)
    if len(txAntSet)==0:
        ret=1
    else:
        ret=len(txAntSet)
    return ret

def getElCritDataList(ptu2InputSetup,antennaList,angleSetupList,confTablesList):
    if ptu2InputSetup['elDiagCriteria']==1:
        angleSetupData=[]
        for antIdx in range(len(antennaList)):
            for angleSetupIdx in range(len(angleSetupList)):
                if angleSetupList[angleSetupIdx]["antennaType"]==antennaList[antIdx]:
                    angleSetupData.append(angleSetupList[angleSetupIdx])
    # To be continued on Monday ;D
    return  angleSetupData

def getAngSetupListPerAntennaType(antennaType,angleSetupList):
    outputList=[]
    for i in range(len(angleSetupList)):
        if angleSetupList[i].antennaType == antennaType:
            outputList.append(angleSetupList[i])
    return outputList



import os #verzeichnise durchsuchen
import fnmatch # dateien filtern (suchmuster)
import io  #datei parsen
import operator


#testpathname = r'//fs9\MDSMS0/Kalibrierung/Kalibrierung_USM/Results/Typ_0x2CXXXX_UMRR-11'
#testpathname = r'\\fs9\MDSMS0\Kalibrierung\Kalibrierung_USM\Results\Typ_0x2CXXXX_UMRR-11\2018\intern_15xUMRR-110004-2C3800Jun2018'
#fileslisttemp = os.listdir(testpathname)  <- nur aktuelles verzeichnis

#for root, dirs, files in os.walk(testpathname):
#   for dir in dirs:
#        print('root: ' + root + ' dir: ' +dir)


class ProtocolInfoClass:
    def __init__(self):
        self.bFileEmpty      = False  # eol found empty protocol file
        self.bErrorOccured   = False # eol error found in protocol file
        self.bAbortCriterion = False # eol abbort criterion found in protocol file
        self.bLimitExceeded  = False  # eol abbort criterion found in protocol file
        self.PathName        = ''   # file path
        self.dateAndTime     = ''  # date and time
        self.productSerial   = ''# product serial number
        self.rf              = '' # rfboard number
        self.dsp             = '' # dsp board number




    # eventuell noch in extra file parse/filter klasse?
    # get filtered file list, filtered by matching search pattern
    @staticmethod
    def getfiltered_ptu2filelist(pathname, searchpattern):
        filteredfilescompletelist = []
        for root, dirs, files in os.walk(pathname):
            filteredlisttemp =  fnmatch.filter(files, searchpattern) #liste der passenden dateien im jeweiligen unterferzeichnis
            #print(filteredlisttemp)
            for filteredfile in filteredlisttemp: # fuege nun pfad + dateinamen zusammen fuer den kompletten dateipfad
                path = root + '/' + filteredfile
                filteredfilescompletelist.append(path)

        #print(filteredfilescompletelist)
        return filteredfilescompletelist

    # # eventuell noch in extra file parse/filter klasse?
    # # get filtered file list, filtered by matching search pattern
    @staticmethod
    def getfiltered_txantdiagramfilelist(pathname, protocolIdentifier):
        filteredfilescompletelist = []
        for root, dirs, files in os.walk(pathname):
            #for file in files:
                #print('root: ' + root + ' file: ' + file)
            for file in files:
                if fnmatch.fnmatch(file,'*'+ protocolIdentifier+'*') and fnmatch.fnmatch(file, '*TxAnt*') and fnmatch.fnmatch(file, '*Diagram*') and  fnmatch.fnmatch(file, '*.txt*') and not fnmatch.fnmatch(file, '*PTU2*'):
                    path=root +'/' + file
                    filteredfilescompletelist.append(path)
                # filteredlisttemp =  fnmatch.filter(files,searchpattern) #liste der passenden dateien im jeweiligen unterferzeichnis
                #print(filteredlisttemp)

            # for filteredfile in filteredlisttemp: # fuege nun pfad + dateinamen zusammen fuer den kompletten dateipfad
            #     filteredfilescompletelist.append(os.path.join(root, filteredfile))


        return filteredfilescompletelist

    # get all lines of this file as strings
    @staticmethod
    def getFileLines(fileNamePath):
        errorlisttemp = []
        content = []
        # Try to read the file
        try:
            f = io.open(fileNamePath, encoding='windows-1252', errors='ignore')

        # Catch an exception
        except IOError as ioErr:
            # todo: den io error noch ausgeben in fehlerlist
            errorlisttemp.append('IO Error File with pathname '+ fileNamePath +  ' does not exist? / Insert a valid file path.')
        else:
            # Read the lines of the file and remove all white spaces
            with f:
                content = f.readlines()
            # strip erstmal and dieser stelle weglassen
            #content = [line.strip() for line in content] # leer zeichen am anfang und ende entfernen
        return content,errorlisttemp


    @staticmethod
    def getPreParseInfo(filelineslist):
        strEOLNoErrorString          = 'E=0;'
        strEOLNoAbortCriterionString = 'K=0;'
        strLimitExceeded             = 'M:'
        dateenum                     = 'Date:'
        timeenum                     = 'Time:'
        prodSerialenum               = 'Product-Serial:'
        rfenum                       = 'RF-Serial:'
        dspenum                      = 'Dsp-Serial:'
        # Initilizing variables
        date            = ''
        dateTime        = ''
        prodSerial      = ''
        rf              = ''
        dsp             = ''
        strHeader       = 'Measurment Durations:'
        bErrorOccured   = True
        bAbortCriterion = True
        bLimitExceeded  = False

        bFileEmpty = len(filelineslist) == 0
        if(bFileEmpty): # datei ist leer -> alle anderen werte auf false
            bErrorOccured = False
            bAbortCriterion = False
            bLimitExceeded = False
        
        for line in filelineslist:
            strTemp = line.strip() #alle leerzeichen am anfang und ende entfernen
            if strTemp.startswith(strHeader): # header zeile erreicht -> suche stoppen
                break
            strTemp = line.replace(' ','') #alle leerzeichen loeschen
            if strTemp.startswith(strEOLNoErrorString):
                bErrorOccured = False
            if strTemp.startswith(strEOLNoAbortCriterionString):
                bAbortCriterion = False
            if strTemp.startswith(strLimitExceeded): # achtung hier ist die suchlogik anders als bei error und abbruchkriterium ( nur wenn  'M:' gefunden wird -> dann ueberschreitung
                bLimitExceeded = True
            if strTemp.startswith(dateenum): # parse date
                linelist = strTemp.split(';')
                if len(linelist) > 1:
                    date = linelist[1]
            if strTemp.startswith(timeenum): # parse time
                linelist = strTemp.split(';')
                if len(linelist) > 1:
                    dateTime = str(date) +'-' + str(linelist[1])
            if strTemp.startswith(prodSerialenum): # parse product serial number
                linelist = strTemp.split(';')
                if len(linelist) > 1:
                    prodSerial = linelist[1]
            if strTemp.startswith(rfenum): # parse rf number 
                linelist = strTemp.split(';')
                if len(linelist) > 1:
                    rf = linelist[1]
            if strTemp.startswith(dspenum): # parse dsp number
                linelist = strTemp.split(';')
                if len(linelist) > 1:
                    dsp = linelist[1]

        return bFileEmpty, bErrorOccured, bAbortCriterion, bLimitExceeded, dateTime, prodSerial, rf, dsp

    # alle protokolldatei im uebergebenen dateipfad und allen unterverzeichnissen
    @staticmethod
    def GetAllProtocolsList(pathname):
        return ProtocolInfoClass.getfiltered_ptu2filelist(pathname, '*proto*.txt') # dateilist gefiltert nach '*proto*.txt'


    # 'vorparsen' der protokolle: information ob fehler, abbruch kriterium oder messgrenzenueberschreitung aufgetreten ist
    @staticmethod
    def PreParseProtcols(protocolpathlist):
        ProtocolPreParseInfosList = []
        for protocolfilepath in protocolpathlist:
            linelist, errorlist = ProtocolInfoClass.getFileLines(protocolfilepath)
            if (len(errorlist) > 0):
                print(errorlist)  # noch durch logger ersetzen
            else:
                # pre parse nach error, abbruchkrit, limit ueberschritten
                # print('lines ' + str(len(linelist)) + ' in ' + protocolfilepath)
                cProtocolInfoTemp = ProtocolInfoClass()
                cProtocolInfoTemp.bFileEmpty, cProtocolInfoTemp.bErrorOccured, cProtocolInfoTemp.bAbortCriterion, cProtocolInfoTemp.bLimitExceeded, \
                    cProtocolInfoTemp.dateAndTime, cProtocolInfoTemp.productSerial, cProtocolInfoTemp.rf, cProtocolInfoTemp.dsp  = ProtocolInfoClass.getPreParseInfo(linelist)
                cProtocolInfoTemp.PathName = protocolfilepath
                ProtocolPreParseInfosList.append(cProtocolInfoTemp)

        bDebug = True
        if (bDebug):
            # print('found ' + str(len(ProtocolPreParseInfosList)) + ' protocols')
            emptyprotocolslist = []
            errorprotocolslist = []
            abbortcritprotocolslist = []
            LimitExceededprotocolslist = []
            for cProtcolPreInfo in ProtocolPreParseInfosList:
                if (cProtcolPreInfo.bFileEmpty):
                    emptyprotocolslist.append(cProtcolPreInfo.PathName)
                if (cProtcolPreInfo.bErrorOccured):
                    errorprotocolslist.append(cProtcolPreInfo.PathName)
                if (cProtcolPreInfo.bAbortCriterion):
                    abbortcritprotocolslist.append(cProtcolPreInfo.PathName)
                if (cProtcolPreInfo.bLimitExceeded):
                    LimitExceededprotocolslist.append(cProtcolPreInfo.PathName)
            # print('empty file in ' + str(len(emptyprotocolslist)) + ' of ' + str(len(ProtocolPreParseInfosList)))
            # print(emptyprotocolslist)
            # print('----------------')
            # print('errors in '+ str(len(errorprotocolslist))+ ' of ' + str(len(ProtocolPreParseInfosList)))
            # print(errorprotocolslist)
            # print('abbort criterions in '+ str(len(abbortcritprotocolslist))+ ' of ' + str(len(ProtocolPreParseInfosList)))
            # print(abbortcritprotocolslist)
            # print('----------------')
            # print('limits exceeded in '+ str(len(LimitExceededprotocolslist))+ ' of ' + str(len(ProtocolPreParseInfosList)))
            # print(abbortcritprotocolslist)
            # print('----------------')
        return ProtocolPreParseInfosList, emptyprotocolslist, errorprotocolslist, abbortcritprotocolslist, LimitExceededprotocolslist


    @staticmethod
    def FilterProtocols(ProtocolPreParseInfosList, bUseFilesWithError, bUseFilesAbbortCriterion, bUseFilesWIthLimitsExceeded):
        filteredProtocolsList    = []
        filtEmptyProtList        = []
        filteredOutProtocolsList = []

        errCrit     = 0
        abbCrit     = 0
        fileLimCrit = 0

        for cProtocolPreParseInfo in ProtocolPreParseInfosList:

            # # wenn keine der ausschluss bedingungen zutrifft -> datei zur 'akzeptiert' list hinzu
            if not (((not bUseFilesWithError) and cProtocolPreParseInfo.bErrorOccured)
                    or ((not bUseFilesAbbortCriterion) and cProtocolPreParseInfo.bAbortCriterion)
                    or ((not bUseFilesWIthLimitsExceeded) and cProtocolPreParseInfo.bLimitExceeded)):
                if not cProtocolPreParseInfo.bFileEmpty:
                    filteredProtocolsList.append(cProtocolPreParseInfo)
            else:
                filteredOutProtocolsList.append(cProtocolPreParseInfo)


        # Sort the list base on date and DSP number
        filteredProtocolsList = sorted(filteredProtocolsList, key=operator.attrgetter('dateAndTime'))
        filteredProtocolsList = sorted(filteredProtocolsList, key=operator.attrgetter('dsp'))
        filteredOutProtocolsList = sorted(filteredOutProtocolsList, key=operator.attrgetter('dateAndTime'))
        filteredOutProtocolsList = sorted(filteredOutProtocolsList, key=operator.attrgetter('dsp'))
        # return the list
        return filteredProtocolsList, filteredOutProtocolsList

    @staticmethod
    def FilterNewProtocols(acceptedProtocolsList):
        filteredNewList = []
        # Sort the list base on date and DSP number
        # acceptedProtocolsList = sorted(acceptedProtocolsList, key=operator.attrgetter('dateAndTime'))
        # acceptedProtocolsList = sorted(acceptedProtocolsList, key=operator.attrgetter('dsp'))
        # sorting
        for idx in range(1, len(acceptedProtocolsList)):
            # Compare serial number
            if acceptedProtocolsList[idx-1].productSerial == acceptedProtocolsList[idx].productSerial:
                # Compare rf and dsp board 
                if acceptedProtocolsList[idx-1].dsp == acceptedProtocolsList[idx].dsp and \
                    acceptedProtocolsList[idx-1].rf == acceptedProtocolsList[idx].rf:
                        # Compare date and time
                        date1 = acceptedProtocolsList[idx-1].dateAndTime
                        date2 = acceptedProtocolsList[idx].dateAndTime
                        if date1 > date2:
                            filteredNewList.append(acceptedProtocolsList[idx-1])
                        else:
                            if idx == len(acceptedProtocolsList)-1:
                                filteredNewList.append(acceptedProtocolsList[idx])
                else:
                    filteredNewList.append(acceptedProtocolsList[idx-1])
                    if idx == len(acceptedProtocolsList)-1:
                        filteredNewList.append(acceptedProtocolsList[idx])
            else:
                filteredNewList.append(acceptedProtocolsList[idx-1])
                if idx == len(acceptedProtocolsList)-1:
                    filteredNewList.append(acceptedProtocolsList[idx])
        
        # If there is only one protocol just use that protocol
        if len(filteredNewList) == 0:
            filteredNewList = acceptedProtocolsList
            
        return filteredNewList

#print('empty file in ' + str(len(emptyprotocolslist)) + ' of ' + str(len(protocolslist)))
#print(emptyprotocolslist)
#print('----------------')
#print('errors in '+ str(len(errorprotocolslist))+ ' of ' + str(len(protocolslist)))
#print(errorprotocolslist)
#print('abbort criterions in '+ str(len(abbortcritprotocolslist))+ ' of ' + str(len(protocolslist)))
#print(abbortcritprotocolslist)
#print('----------------')
#print('limits exceeded in '+ str(len(LimitExceededprotocolslist))+ ' of ' + str(len(protocolslist)))
#print(abbortcritprotocolslist)
#print('----------------')
#print('EOL Ok in '+ str(len(Okprotocolslist))+ ' of ' + str(len(protocolslist)))
#print(Okprotocolslist)
#print('----------------')
#print('end')
#return Okprotocolslist

